#include "global.h" 
#include "functions.h"
#include "opcode.h"
#include "regs.h"
#include "flags.h"
#include "fltconst.h"
#include "blocks.h"

#define MAX_NR_DESCRIPTORS 10000

static int number_of_address_descriptors;

#define ABBL_WORKING(i) ABBL_U2(i).doubleword.d1
#define AEDG_WORKING(i) AEDG_U1(i).doubleword.d1
#define AEDG_PROP_REGS(i) AEDG_U2(i).quadword

static INT64 address;

/********************************************************
 print an address descriptor
 ********************************************************/

extern void PrintAddressDescriptor(ADDRESSDESCRIPTOR * ad)
{
  int i;
  printf("regs      : %lx\n",ad->regs);
  printf("const regs: %lx\n",ad->const_regs);
  for (i=0;i<32;i++)
    if ((1<<i) & ad->const_regs)
      printf("%2d : %lx\n",i,ad->constants[i]); 
}

/********************************************************
 set the registers for which addresses must be propagated
 over the edges
 ********************************************************/
extern void SetEdgPropRegs()
{
  INDEX iedg;

  FOREACH_LIVE_EDG(iedg)
    {

      // naar hell-lite

      if (AEDG_TYPE(iedg)==ET_COMPENSATE && ABBL_FUN(AEDG_TAIL(iedg))==AFUN_PSEUDO_HELL_LITE)
	AEDG_PROP_REGS(iedg) = REGS_IRET | REGS_FRET;

      // naar pal

      else if (AEDG_TYPE(iedg)==ET_FUN_CALL && ABBL_FUN(AEDG_TAIL(iedg))==AFUN_PSEUDO_PAL)
	{
	  AEDG_PROP_REGS(iedg) = REGS_IARG | REGS_IRET;
	  AEDG_PROP_REGS(iedg+1) = REGS_IRET;
	}

      else if (AEDG_TYPE(iedg)==ET_FUN_LINK && ABBL_FUN(AEDG_TAIL(ABBL_EDG_CALL(AEDG_HEAD(iedg))))==AFUN_PSEUDO_PAL)
	AEDG_PROP_REGS(iedg) = (REGS_PROC_INT_PRESERVED | REGS_PROC_FLT_PRESERVED) & AbblRegsLiveIn(AEDG_TAIL(iedg));

      // naar hell

      else if (AEDG_TYPE(iedg)==ET_FUN_CALL && ABBL_FUN(AEDG_TAIL(iedg))==AFUN_PSEUDO_HELL)
	{
	  AEDG_PROP_REGS(iedg) = REGS_IARG;
	  AEDG_PROP_REGS(iedg+1) = REGS_IRET;
	}

      else if (AEDG_TYPE(iedg)==ET_FUN_LINK && ABBL_FUN(AEDG_TAIL(ABBL_EDG_CALL(AEDG_HEAD(iedg))))==AFUN_PSEUDO_HELL)
	AEDG_PROP_REGS(iedg) = (REGS_PROC_INT_PRESERVED | REGS_PROC_FLT_PRESERVED) & AbblRegsLiveIn(AEDG_TAIL(iedg));

      // naar call

      else if (AEDG_TYPE(iedg)==ET_FUN_CALL && ABBL_FUN(AEDG_TAIL(iedg))==AFUN_PSEUDO_CALL)
	{
	  AEDG_PROP_REGS(iedg) = REGS_IARG | REGS_FARG;
	  AEDG_PROP_REGS(iedg+1) = REGS_IRET | REGS_FRET;
	}
      else if (AEDG_TYPE(iedg)==ET_FUN_LINK && ABBL_FUN(AEDG_TAIL(ABBL_EDG_CALL(AEDG_HEAD(iedg))))==AFUN_PSEUDO_CALL)
	AEDG_PROP_REGS(iedg) = (REGS_PROC_INT_PRESERVED | REGS_PROC_FLT_PRESERVED) & AbblRegsLiveIn(AEDG_TAIL(iedg));
      
      // gewone

      else if (AEDG_TYPE(iedg)==ET_FUN_CALL)
	AEDG_PROP_REGS(iedg)= (~AFUN_REGS_SAVED(ABBL_FUN(AEDG_TAIL(iedg)))) & AbblRegsLiveIn(AEDG_TAIL(iedg));

      else if (AEDG_TYPE(iedg)==ET_FUN_LINK)
	AEDG_PROP_REGS(iedg)= (AFUN_REGS_SAVED(ABBL_FUN(AEDG_TAIL(ABBL_EDG_CALL(AEDG_HEAD(iedg)))))) & AbblRegsLiveIn(AEDG_TAIL(iedg));

      else if (AEDG_TYPE(iedg)==ET_FUN_RETURN && ABBL_FUN(AEDG_HEAD(iedg)) != AFUN_PSEUDO_PAL && ABBL_FUN(AEDG_HEAD(iedg)) != AFUN_PSEUDO_CALL)
	AEDG_PROP_REGS(iedg)= (~AFUN_REGS_SAVED(ABBL_FUN(AEDG_HEAD(iedg)))) & AbblRegsLiveIn(AEDG_TAIL(iedg));

      else
	AEDG_PROP_REGS(iedg)=AbblRegsLiveIn(AEDG_TAIL(iedg));
    }
}

/********************************************************
 initialize the marking of blocks and edges for
 ainsmarkuses
 ********************************************************/
extern void InitMarking()
{
  INDEX ibbl;
  INDEX iedg;

  FOREACH_BBL(ibbl)
    ABBL_WORKING(ibbl) = -1;
  
  FOREACH_EDG(iedg)
    AEDG_WORKING(iedg) = -1;
}

/********************************************************
 mark the return edges over which addresses generated
 by an instructions can be propagated (to be able to
 do some kind of context-sensitive propagation of addresses
 ********************************************************/

extern void AfunMarkReturnEdgesRecursive(INDEX ifun,INDEX def, int hola)
{
  INDEX iedg;

  if (AFUN_IS_PSEUDO(ifun)) return;

  FOREACH_BBL_SUCC(AFUN_BBL_LAST(ifun),iedg)
    {
      if (AEDG_TYPE(iedg)==ET_COMPENSATE && ABBL_FUN(AEDG_TAIL(iedg))==AFUN_PSEUDO_HELL_LITE && ABBL_FUN(AINS_BBL(def))!=ifun) continue;
      if (AEDG_TYPE(iedg)==ET_COMPENSATE && ABBL_FUN(AEDG_TAIL(iedg))==AFUN_PSEUDO_HELL && ABBL_FUN(AINS_BBL(def))!=ifun) continue;
      if (AEDG_WORKING(iedg)==def) continue;
      //      if (hola)      printf("setting %c edge from fun %d to bbl %d in fun %d\n",AEDG_TYPE(iedg),ifun,AEDG_TAIL(iedg),ABBL_FUN(AEDG_TAIL(iedg)));
      AEDG_WORKING(iedg)=def;     
      AfunMarkReturnEdgesRecursive(ABBL_FUN(AEDG_TAIL(iedg)),def,hola);
    }
}

/********************************************************
 part of the garbage collector of propagated addresses
 free's all address descriptors
 ********************************************************/
extern void ResetAddressCalc()
{
  INDEX ibbl;
  FOREACH_BBL(ibbl)
    {
      if (ABBL_ADDRESS(ibbl))
	free(ABBL_ADDRESS(ibbl));
      ABBL_ADDRESS(ibbl)=0L;
    }
  number_of_address_descriptors=0;
}

/********************************************************
 part of the garbage collector: free address descriptors
 coming from previous address propagations
 ********************************************************/
extern void CleanUpAddressDescriptors(INDEX inst)
{
  INDEX ibbl;
  printf("cleanup\n");
  FOREACH_BBL(ibbl)
    if (ABBL_ADDRESS(ibbl))
      if ((ABBL_ADDRESS(ibbl)->inst)!=inst)
	{
	  free(ABBL_ADDRESS(ibbl));
	  ABBL_ADDRESS(ibbl)=0L;
	  number_of_address_descriptors--;
	}
}
/********************************************************
 something like joinvaluespec, but with garbage collector
 in mind: sometimes memory has to be provided first
 ********************************************************/

extern BOOL AbblJoinAddressDescriptorRegs(INDEX bbl, ADDRESSDESCRIPTOR* ad,REGSET regs)
{

  if (ABBL_ADDRESS(bbl)==0)
    {
      // create new descriptor
      
      // first do garbage collection if necessary
      
      if (MAX_NR_DESCRIPTORS == number_of_address_descriptors)
	CleanUpAddressDescriptors(ad->inst);
      
      // now allocate space
      ABBL_ADDRESS(bbl) = (ADDRESSDESCRIPTOR*) malloc (sizeof(ADDRESSDESCRIPTOR));
      // copy the descriptor;
      memcpy(ABBL_ADDRESS(bbl),ad,sizeof(ADDRESSDESCRIPTOR));

      ABBL_ADDRESS(bbl)->regs &= regs;
      ABBL_ADDRESS(bbl)->const_regs &= regs;

      if (ABBL_ADDRESS(bbl)->regs)
	return 1;
      else
	return 0;
    }
  
  if (ABBL_ADDRESS(bbl)->inst == ad->inst)
    {
      // join the descriptor
      
      ADDRESSDESCRIPTOR* new_ad;
      REG reg;
      
      new_ad = (ADDRESSDESCRIPTOR*) malloc (sizeof(ADDRESSDESCRIPTOR));
      memcpy(new_ad,ABBL_ADDRESS(bbl),sizeof(ADDRESSDESCRIPTOR));
  
      // find the join of regs that contain addresses
      
      new_ad->regs  = (ad->regs & regs) | ABBL_ADDRESS(bbl)->regs;

      // find the join of regs with constant addresses;

      new_ad->const_regs = (ad->const_regs & ad->regs & regs) & ABBL_ADDRESS(bbl)->const_regs;

      REGSET_FOREACH_REG(new_ad->const_regs,reg)
	if ((ABBL_ADDRESS(bbl)->constants[reg])!=(ad->constants[reg]))
	  new_ad->const_regs &= ~(REG2BIT(reg));
      
      if (bcmp(new_ad,ABBL_ADDRESS(bbl),sizeof(ADDRESSDESCRIPTOR)))
	{
	  free(ABBL_ADDRESS(bbl));
	  ABBL_ADDRESS(bbl)=new_ad;
	  return 1;
	}
      else
	{
	  free(new_ad);
	  return 0;
	}
    }

  // copy the descriptor
  memcpy(ABBL_ADDRESS(bbl),ad,sizeof(ADDRESSDESCRIPTOR));
  ABBL_ADDRESS(bbl)->regs &= regs;
  ABBL_ADDRESS(bbl)->const_regs &= regs;
  if (ABBL_ADDRESS(bbl)->regs)
    return 1;
  else
    return 0;
}

/********************************************************
 propagate an address descriptor into and over an instruction
 and return whether this instruction does something with data
 ********************************************************/

extern int AinsEvalAddressDescriptor(INDEX iins,ADDRESSDESCRIPTOR* ad,INDEX def)
{
  int function = 0;

  if (AINS_REG_C(iins)!=REG_ZERO && (AINS_FLAGS(iins) & FI_EVAL_C))
    {
      ad->regs       &= ~AinsDef(iins);
      ad->const_regs &= ~AinsDef(iins);
    }
  else if (AinsUse(iins) & ad->regs)
    {
      if (AINS_IS_LOAD(iins))
	{
	  if (!(AINS_FLAGS(iins) & FI_EVAL_B))
	    {
	      if (REG2BIT(AINS_REG_B(iins)) & ad->const_regs)
		{
		  address  = AINS_DATA(iins)+ad->constants[AINS_REG_B(iins)];
		  function = CONST_LOAD;
		}
	      else
		{
		  function = BLOCK_LOAD;
		}
	    }
	  ad->regs       &= ~AinsDef(iins);
	  ad->const_regs &= ~AinsDef(iins);
	}
      else if (AINS_IS_STORE(iins))
	{
	  if (REG2BIT(AINS_REG_A(iins)) & ad->regs)
	    {
	      //	      printf("def %d bbl %d wordt gestored\n",ad->inst,AINS_BBL(ad->inst));
	      //	      PrintIns(stdout,ad->inst);
	      //	      PrintIns(stdout,iins);
	      function = BLOCK_STORE | BLOCK_LOAD;
	    }
	  else if (!(AINS_FLAGS(iins) & FI_EVAL_B))
	    {
	      if (REG2BIT(AINS_REG_B(iins)) & ad->const_regs)
		{
		  address  = AINS_DATA(iins)+ad->constants[AINS_REG_B(iins)];
		  function = CONST_STORE;
		}
	      else
		{
		  //		  printf("def %d bbl %d heeft dirty store\n",ad->inst,AINS_BBL(ad->inst));
		  //		  PrintIns(stdout,ad->inst);
		  //		  PrintIns(stdout,iins);
		  function = BLOCK_STORE;
		}
	    }
	}
      else if ( AINS_TYPE(iins) != IT_IBR && AINS_TYPE(iins) != IT_JMP && AINS_TYPE(iins) != IT_FBR && AINS_TYPE(iins) != IT_ICP && AINS_TYPE(iins) != IT_FCP && !AINS_IS_LOAD(iins) && !(AINS_CODE(iins)==I_and && AINS_REG_B(iins)==REG_LIT) && !((AINS_TYPE(iins) == IT_SFT) && !((REG2BIT(AINS_REG_A(iins)) & ad->regs))) && !((AINS_TYPE(iins) == IT_ICM) && !((REG2BIT(AINS_REG_B(iins)) & ad->regs)) && !((REG2BIT(AINS_REG_C(iins)) & ad->regs))) )
	{
	  VALUESPEC* spec = (VALUESPEC*) malloc (sizeof(VALUESPEC));
	  int i;
	  INDEX reg;
	  
	  spec->bot = 0L;
	  spec->top = 0L;

	  for (i=0;i<31;i++)
	    {
	      reg = REG2BIT(i);
	      if (ad->const_regs & reg)
		{
		  spec->value[i].i = ad->constants[i];
		  spec->bot &= ~reg;
		  spec->top &= ~reg;
		}
	      else if (ad->regs & reg)
		{
		  spec->value[i].i = 0;
		  spec->bot &= ~reg;
		  spec->top |= reg;
		}
	      else
		{
		  spec->value[i].i = 0;
		  spec->bot |=  reg;
		  spec->top &= ~reg;
		}
	    }
	  spec->value[31].i = 0L;
	  
	  if (AINS_REG_A(iins)!=REG_LIT && (AINS_FLAGS(iins) & FI_EVAL_A))
	    spec->value[AINS_REG_A(iins)].i = AINS_A_EVAL(iins);
	  if (AINS_REG_B(iins)!=REG_LIT && (AINS_FLAGS(iins) & FI_EVAL_B))
	    spec->value[AINS_REG_B(iins)].i = AINS_B_EVAL(iins);

	  AinsEvalValueSpecNew(iins,spec);

	  if (!((spec->bot | spec->top ) & REG2BIT(AINS_REG_C(iins))))
	    {
	      ad->regs       |= AinsDef(iins);
	      ad->const_regs |= AinsDef(iins);
	      ad->constants[AINS_REG_C(iins)] = spec->value[AINS_REG_C(iins)].i;
	    }
	  else 
	    {
	      ad->regs       |= AinsDef(iins);
	      ad->const_regs &= ~AinsDef(iins);
	    }
	  free(spec);
	}
    }
  else
    {
      ad->regs       &= ~AinsDef(iins);
      ad->const_regs &= ~AinsDef(iins);
    }
  
  return function;
}

/********************************************************
 mark all the uses of an address generated by an instructions
 and adapt the data map accordingly
 ********************************************************/


extern void AinsMarkUses(INDEX def)
{
  INDEX ifun;
  INDEX iins;
  INDEX ibbl;
  INDEX iedg;
  BOOL change;
  BOOL change2;
  int function;
  int flags = 0;

  ADDRESSDESCRIPTOR* ad = (ADDRESSDESCRIPTOR*) malloc(sizeof(ADDRESSDESCRIPTOR));

  AfunMarkReturnEdgesRecursive(ABBL_FUN(AINS_BBL(def)),def,def==75217);

  iins = AINS_NEXT(def);

  ad->const_regs = ad->regs      = AinsDef(def);
  ad->inst                       = def;
  ad->constants[AINS_REG_C(def)] = AINS_C_EVAL(def);
  
  while (VALID(iins))
    {
      if (AINS_FLAGS(iins) & F_KILLED)
	{
	  iins=AINS_NEXT(iins);
	  continue;
	}
      
      flags |= (function = AinsEvalAddressDescriptor(iins,ad,def));
      
      if (function)
	{
	  if (function & CONST_LOAD)
	    AdaptMap(CONST_LOAD,address);
	  if (function & CONST_STORE)
	    AdaptMap(CONST_STORE,address);
	  if (function & BLOCK_LOAD)
	    AdaptMap(BLOCK_LOAD,AINS_C_EVAL(def));
	  if (function & BLOCK_STORE)
	    {
	      AdaptMap(BLOCK_STORE,AINS_C_EVAL(def));
	    }
	  if ((flags & BLOCK_STORE) && (flags & BLOCK_LOAD))
	    {
	      free(ad);
	      return;
	    }
	}
      iins=AINS_NEXT(iins);
    }
  
  ad->regs &= ABBL_OUT(AINS_BBL(def));
  
  if (!(ad->regs))
    {
      free(ad);
      return;
    }

  FOREACH_LIVE_FUN(ifun)
    AFUN_UNMARK(ifun);

  FOREACH_LIVE_BBL(ibbl)
    ABBL_UNMARK(ibbl);

  ibbl = AINS_BBL(def);

  AFUN_MARK(ABBL_FUN(ibbl));

  FOREACH_BBL_SUCC(ibbl,iedg)
    {
      if (AEDG_TYPE(iedg)==ET_FUN_LINK) continue;
      if (AEDG_IS_INTER(iedg))
	{
	  if (AEDG_TYPE(iedg)==ET_FUN_CALL)// && !AFUN_IS_PSEUDO(ABBL_FUN(AEDG_TAIL(iedg))))
	    {
	      INDEX link_edge;
	      AEDG_WORKING(iedg+1) = def;
	      if (AbblJoinAddressDescriptorRegs(AEDG_TAIL(iedg),ad,AEDG_PROP_REGS(iedg)))
		{
		  AFUN_MARK(ABBL_FUN(AEDG_TAIL(iedg)));
		  ABBL_MARK(AEDG_TAIL(iedg));
		  ABBL_WORKING(AEDG_TAIL(iedg)) = def;
		  change = 1;
		}
	      link_edge = ABBL_EDG_LINK(ibbl);
	      if (AbblJoinAddressDescriptorRegs(AEDG_TAIL(link_edge),ad,AEDG_PROP_REGS(link_edge)))
		{
		  ABBL_MARK(AEDG_TAIL(link_edge));
		  AFUN_MARK(ABBL_FUN(AEDG_TAIL(link_edge)));
		  ABBL_WORKING(AEDG_TAIL(link_edge)) = def;
		  change = 1;
		}
	      
	      continue;
	    }	 
	  
	  if (AEDG_TYPE(iedg)==ET_FUN_RETURN && AEDG_WORKING(iedg) != def)
	    continue;
	  if (AEDG_IS_INTER(iedg) && AEDG_TYPE(iedg)!=ET_FUN_CALL && AEDG_TYPE(iedg)!=ET_FUN_RETURN)
	    AfunMarkReturnEdgesRecursive(ABBL_FUN(AEDG_TAIL(iedg)),def,def==75217);
	  if (AEDG_TYPE(iedg)==ET_FUN_RETURN)
	    {
	      if (AbblJoinAddressDescriptorRegs(AEDG_TAIL(iedg),ad,AEDG_PROP_REGS(iedg)))
		{
		  AFUN_MARK(ABBL_FUN(AEDG_TAIL(iedg)));
		  ABBL_MARK(AEDG_TAIL(iedg));
		  ABBL_WORKING(AEDG_TAIL(iedg)) = def;
		  change = 1;
		}
	      continue;
	    }
	  if (AbblJoinAddressDescriptorRegs(AEDG_TAIL(iedg),ad,AEDG_PROP_REGS(iedg)))
	    {
	      ABBL_MARK(AEDG_TAIL(iedg));
	    AFUN_MARK(ABBL_FUN(AEDG_TAIL(iedg)));
	    ABBL_WORKING(AEDG_TAIL(iedg)) = def;
	    change = 1;
	    }
	}
      else
	if (AbblJoinAddressDescriptorRegs(AEDG_TAIL(iedg),ad,AEDG_PROP_REGS(iedg)))
	  {
	    ABBL_MARK(AEDG_TAIL(iedg));
	    AFUN_MARK(ABBL_FUN(AEDG_TAIL(iedg)));
	    ABBL_WORKING(AEDG_TAIL(iedg)) = def;
	    change = 1;
	  }
    }
  
  change = 1;
  
  while (change)
    {
      change = 0;
      
      FOREACH_LIVE_FUN(ifun)
	{
	  if (!AFUN_IS_MARKED(ifun)) continue;
	  
	  AFUN_UNMARK(ifun);

	  if (AFUN_IS_PSEUDO(ifun))
	    {
	      AdaptMap(BLOCK_LOAD,AINS_C_EVAL(def));
	      AdaptMap(BLOCK_STORE,AINS_C_EVAL(def));
	      free(ad);
	      return;
	    }
	  
	  change2 = 1;
	  
	  while (change2)
	    {
	      change2 = 0;

	      FOREACH_FUN_BBL(ifun,ibbl)
		{
		  if (ABBL_WORKING(ibbl) != def) continue;
		  if (!ABBL_IS_MARKED(ibbl)) continue;
		  ABBL_UNMARK(ibbl);

		  ABBL_WORKING(ibbl) = -1;
		  
		  memcpy(ad,ABBL_ADDRESS(ibbl),sizeof(ADDRESSDESCRIPTOR));
		  
		  FOREACH_BBL_INS(ibbl,iins)
		    {
		      if (AINS_FLAGS(iins) & F_KILLED)
			continue;
		      
		      flags |= function = AinsEvalAddressDescriptor(iins,ad,def);
		      
		      if (function)
			{
			  if (function & CONST_LOAD)
			    AdaptMap(CONST_LOAD,address);
			  if (function & CONST_STORE)
			    AdaptMap(CONST_STORE,address);
			  if (function & BLOCK_LOAD)
			    AdaptMap(BLOCK_LOAD,AINS_C_EVAL(def));
			  if (function & BLOCK_STORE)
			    {
			      AdaptMap(BLOCK_STORE,AINS_C_EVAL(def)); 
			    }
			  if ((flags & BLOCK_STORE) && (flags & BLOCK_LOAD))
			    {
			      free(ad);
			      return;
			    }
			}
		    }
		  
		  ad->regs &= ABBL_OUT(ibbl);
		  
		  if (!(ad->regs)) continue;
		  
		  FOREACH_BBL_SUCC(ibbl,iedg)
		    {
		      if (AEDG_TYPE(iedg)==ET_FUN_LINK) continue;
		      if (AEDG_IS_INTER(iedg))
			{
			  if (AEDG_TYPE(iedg)==ET_FUN_CALL)// && !AFUN_IS_PSEUDO(ABBL_FUN(AEDG_TAIL(iedg))))
			    {
			      INDEX link_edge;
			      AEDG_WORKING(iedg+1) = def;
			      if (AbblJoinAddressDescriptorRegs(AEDG_TAIL(iedg),ad,AEDG_PROP_REGS(iedg)))
				{
				  AFUN_MARK(ABBL_FUN(AEDG_TAIL(iedg)));
				  ABBL_MARK(AEDG_TAIL(iedg));		   
				  ABBL_WORKING(AEDG_TAIL(iedg)) = def;
				  change = 1;
				}
			      link_edge = ABBL_EDG_LINK(ibbl);
			      
			      if (AbblJoinAddressDescriptorRegs(AEDG_TAIL(link_edge),ad,AEDG_PROP_REGS(link_edge)))
				{
				  ABBL_MARK(AEDG_TAIL(link_edge));
				  AFUN_MARK(ABBL_FUN(AEDG_TAIL(link_edge)));
				  ABBL_WORKING(AEDG_TAIL(link_edge)) = def;
				  change2 = 1;
				}
			      
			      continue;
			    }	 
			  
			  if (AEDG_TYPE(iedg)==ET_FUN_RETURN && AEDG_WORKING(iedg) != def)
			    continue;
			  if (AEDG_TYPE(iedg)==ET_COMPENSATE && ibbl==AFUN_BBL_LAST(ifun) && ABBL_FUN(AEDG_TAIL(iedg))==AFUN_PSEUDO_HELL_LITE && AEDG_WORKING(iedg) != def)
			    continue;
			  if (AEDG_TYPE(iedg)==ET_COMPENSATE && ibbl==AFUN_BBL_LAST(ifun) && ABBL_FUN(AEDG_TAIL(iedg))==AFUN_PSEUDO_HELL && AEDG_WORKING(iedg) != def)
			    continue;
			  if (AEDG_TYPE(iedg)==ET_COMPENSATE && ibbl==AFUN_BBL_LAST(ifun) && AEDG_WORKING(iedg)!=def)
			    continue;
			  if (AEDG_IS_INTER(iedg) && AEDG_TYPE(iedg)!=ET_FUN_CALL && AEDG_TYPE(iedg)!=ET_FUN_RETURN && ibbl!=AFUN_BBL_LAST(ifun))
			    {
			      AfunMarkReturnEdgesRecursive(ABBL_FUN(AEDG_TAIL(iedg)),def,def==75217);
			    }
			  if (AEDG_TYPE(iedg)==ET_FUN_RETURN)
			    {
			      if (AbblJoinAddressDescriptorRegs(AEDG_TAIL(iedg),ad,AEDG_PROP_REGS(iedg)))
				{
				  ABBL_WORKING(AEDG_TAIL(iedg)) = def;
				  AFUN_MARK(ABBL_FUN(AEDG_TAIL(iedg)));
				  ABBL_MARK(AEDG_TAIL(iedg));
				  change = 1;
				}
			      continue;
			    }
			  if (AbblJoinAddressDescriptorRegs(AEDG_TAIL(iedg),ad,AEDG_PROP_REGS(iedg)))
			    {
			      ABBL_MARK(AEDG_TAIL(iedg));
			      ABBL_WORKING(AEDG_TAIL(iedg)) = def;
			      if (AEDG_IS_INTER(iedg))
				{
				  AFUN_MARK(ABBL_FUN(AEDG_TAIL(iedg)));
				  change = 1;
				}
			      else change2 = 1;
			    }
			}
		      else
			if (AbblJoinAddressDescriptorRegs(AEDG_TAIL(iedg),ad,AEDG_PROP_REGS(iedg)))
			  {
			    ABBL_MARK(AEDG_TAIL(iedg));
			    ABBL_WORKING(AEDG_TAIL(iedg)) = def;
			    change2 = 1;
			  }
		    }
		}
	    }
	}
    }
  free(ad);
  return;
}
