#include "global.h"
#include "flags.h"
#include "functions.h"
#include "regs.h"
#include "opcode.h"

#define SaveRegDispl(x)      (x == 26 ? 0 : 8*(x-8))

#define MIN(x, y)            (x < y ? x : y)

#define FACTOR_BBL_LST(x)    (x->bbl_list)
#define FACTOR_ROOT(x)       (x->root)
#define FACTOR_BBL(x)        (x->bbl)
#define PROCESSED(x)         (x->processed)

/*=======================================================
=======================================================*/

static REG MaxRegNo(REGSET regs);
static REG MinRegNo(REGSET regs);
static NUMBER NumRegs(REGSET regs);
static BOOL AbblBadSave(INDEX ibbl);
static BOOL AbblSavesRA(INDEX ibbl);

REG in_subst[64], out_subst[64];

#define InSubst(i)  in_subst[i]
#define OutSubst(i) out_subst[i]

#define NUM_OPCODES 550
NUMBER opcode_counts[NUM_OPCODES], opcode_list[NUM_OPCODES];
UINT8  opcode_code[NUM_OPCODES];

#define OpcodeCounts(i) opcode_counts[i]
#define OpcodeCode(i) opcode_code[i]
#define Opcode(i) opcode_list[i]

#define NUM_TOP_OPCODES 255
#define NUM_INSTS_ENCODED 8

#define NUM_BUCKETS 512
static INDEX hashtbl[NUM_BUCKETS];

#define HashTbl(i)  hashtbl[i]

static INT64 rotate_table[64];
static BOOL CompareDagBbls(INDEX ibbl,INDEX jbbl);
static BOOL DetermineSubsts(INDEX ibbl,INDEX jbbl);
static void AddMoves(INDEX jbbl);


/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "cmp"

static int cmp(const void *x1,const void *x2)
{
  NUMBER n1 = *(NUMBER *)x1;
  NUMBER n2 = *(NUMBER *)x2;
    
  if ( OpcodeCounts(n1) > OpcodeCounts(n2) ) 
    return -1;
  else if ( OpcodeCounts(n1) < OpcodeCounts(n2) ) 
    return 1;
  else 
    return 0;
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "AllocNode"

extern void *AllocNode(size_t size)
{
  void *x = malloc(size);
  if (!x) {
    fprintf(stderr, "out of memory!\n");
    exit(-1);
  }

  return x;
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "NewFunName"

#define STRBUFSZ 128

extern char *NewFunName(char *root, NUMBER r)
{
  char fname[STRBUFSZ];

  strcpy(fname, root);
  sprintf(&fname[strlen(root)], "%d", r);

  return strdup(fname);
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "FactorSaveRegs"

extern void FactorSaveRegs()
{
  NUMBER benefit[32], maxbenefit;
  int i;
  INDEX ifun, xfun;
  INDEX ibbl, jbbl, sbbl;
  INDEX iedg;
  INDEX iins;
  INDEX saveregbbls[8];
  REGSET regs;
  REGSET regs_to_consider;
  REG r, r0;
  BOOL chg;
  NUMBER nins = 0, ains = 0;
  BOOL verbose = HasOption(PRINT_OPT_VERBOSE, "FactorSaveRegs");
  BOOL xverbose = HasOption(PRINT_OPT_VERBOSE, "FuncNames");

  ComLiveness(stdout);

  FOREACH_LIVE_FUN(ifun) {
    ABBL_UNMARK(AFUN_BBL_FIRST(ifun));
  }    

  do {
    chg = FALSE;

    for (i=0; i<32; i++) benefit[i] = 0;

    FOREACH_LIVE_FUN(ifun) {
      if (AFUN_IS_PSEUDO(ifun)) continue;
      if (ABBL_IS_MARKED(AFUN_BBL_FIRST(ifun))) continue;
      ibbl = AFUN_BBL_FIRST(ifun);
      if (AbblBadSave(ibbl)) continue;
      if (!AbblSavesRA(ibbl)) continue;
      if (NumRegs(AbblSavesRegs(ibbl) & SAVE_REGS) == 0) continue;

      regs = REGSET_DIFF(REGS_INT_BUT_ZERO,AbblRegsLiveIn(ibbl));

      regs_to_consider = AbblSavesRegs(ibbl);
      FOREACH_BBL_INS(ibbl, iins) {
	if (REGSET_IS_EMPTY(regs_to_consider)) break;

	if (REGSET_IS_MEMBER(AINS_REG_A(iins), regs_to_consider)
	    && AINS_CODE(iins) == I_stq
	    && AINS_REG_B(iins) == REG_SP
	    && AINS_DATA(iins) == SaveRegDispl(AINS_REG_A(iins))) {
	  regs_to_consider = REGSET_DIFF(regs_to_consider, REG2BIT(AINS_REG_A(iins)));
	}

	regs = REGSET_DIFF(regs, AinsDef(iins));
      }

      REGSET_FOREACH_REG(regs, r) {
	benefit[r] += NumRegs(AbblSavesRegs(ibbl));
      }
    }

    maxbenefit = -1;
    r0 = INVALID;
    for (i = 0; i < 32; i++) {
      if (benefit[i] > maxbenefit) {
	maxbenefit = benefit[i];
	r0 = i;
      }
    }

    if (maxbenefit < 10) break;  /* we'll be adding 9+1 = 10 instructions */

    benefit[r0] = -1;

    /*
     * create new functions that save the callee save registers and return
     * through register r0.
     */
    xfun = AfunNewFancy();
    AFUN_NAME(xfun) = NewFunName("_ssave", r0);
    AFUN_FLAGS(xfun) = (FF_STK_INVALID|FF_STK_ARG|FF_STK_ALIAS);

    sbbl = AbblNew();
    ABBL_FUN(sbbl) = xfun;
    ABBL_TYPE(sbbl) = BT_RET;
    ABBL_COUNT(sbbl) = 0;
    AbblAppendFancy(sbbl);
  
    iins = AinsNew();
    AinsMakeStq(iins, REG_RA, REG_SP, 0);        /* 0($sp) := $26 */
    AINS_BBL(iins) = sbbl;
    AinsAppend(iins);
    ains++;
    
    iins = AinsNew();
    AinsMakeRet(iins, r0);                   /* ret (r0) */
    AINS_BBL(iins) = sbbl;
    AinsAppend(iins);
    ains++;
  
    AedgLinkFancy( AedgNewFancy(sbbl, sbbl, ET_EXIT_LINK) );

    saveregbbls[0] = sbbl;

    i = 1;
    REGSET_FOREACH_REG(SAVE_REGS, r) {
      xfun = AfunNewFancy();
      AFUN_NAME(xfun) = NewFunName("_ssave", r0);
      AFUN_FLAGS(xfun) = (FF_STK_INVALID|FF_STK_ARG|FF_STK_ALIAS);

      sbbl = AbblNew();
      ABBL_FUN(sbbl) = xfun;
      ABBL_TYPE(sbbl) = BT_NORM;
      ABBL_COUNT(sbbl) = 0;
      AbblAppendFancy(sbbl);
      saveregbbls[i] = sbbl;
      AedgLinkFancy(AedgNewFancy(saveregbbls[i], saveregbbls[i-1], ET_NORM));
      i += 1;
    
      iins = AinsNew();
      AinsMakeStq(iins, r, REG_SP, SaveRegDispl(r));
      AINS_BBL(iins) = sbbl;
      AinsAppend(iins);
      ains++;
    }

    FOREACH_LIVE_FUN(ifun) {
      if (AFUN_IS_PSEUDO(ifun)) continue;
      ibbl = AFUN_BBL_FIRST(ifun);
      if (ABBL_IS_MARKED(ibbl)) continue;
      if (AbblBadSave(ibbl)) continue;
      if (!AbblSavesRA(ibbl)) continue;
      if (NumRegs(AbblSavesRegs(ibbl) & SAVE_REGS) == 0) continue;

      regs = REGSET_DIFF(REGS_INT_BUT_ZERO, AbblRegsLiveIn(ibbl));
      /*
       * When choosing a return register for the function call to
       * the function that saves callee-saved registers, we can't 
       * use any register that is defined by any of the instructions
       * between the start of the block and the last instruction that
       * saves a callee-saved register, so we discard such regs.  Notice
       * that it's OK to use a register that is defined in the block
       * but not until after the last relevant save instruction.
       */
      regs_to_consider = AbblSavesRegs(ibbl);
      FOREACH_BBL_INS(ibbl, iins) {
	if (REGSET_IS_EMPTY(regs_to_consider)) break;

	if (REGSET_IS_MEMBER(AINS_REG_A(iins), regs_to_consider)
	    && AINS_CODE(iins) == I_stq
	    && AINS_REG_B(iins) == REG_SP
	    && AINS_DATA(iins) == SaveRegDispl(AINS_REG_A(iins))) {
	  regs_to_consider = REGSET_DIFF(regs_to_consider, REG2BIT(AINS_REG_A(iins)));
	}

	regs = REGSET_DIFF(regs, AinsDef(iins));
      }

      if (!REGSET_IS_MEMBER(r0, regs)) {
	if (verbose) {
	  printf("\tFactorSaveRegs: no free regs: bbl %d [%d regs saved]\n", 
		 ibbl, NumRegs(AbblSavesRegs(ibbl)));
	}
	continue;
      }
      ABBL_MARK(ibbl);
      chg = TRUE;
      r = MaxRegNo( (AbblSavesRegs(ibbl) & SAVE_REGS) );
      regs_to_consider = AbblSavesRegs(ibbl);

      jbbl = AbblNew();
      ABBL_FUN(jbbl)  = ABBL_FUN(ibbl);
      ABBL_TYPE(jbbl) = ABBL_TYPE(ibbl);
      ABBL_COUNT(jbbl) = ABBL_COUNT(ibbl);
      AbblAppendFancy(jbbl);

      if (verbose) {
	printf("@@@ FactorSaveRegs: ibbl %d [fun %d, %s] -> jbbl %d\n", 
	       ibbl, ABBL_FUN(ibbl), AFUN_NAME(ABBL_FUN(ibbl)), jbbl);
      }
      /*
       * First, walk down the list of instructions till we find the first
       * instruction that stores a callee-saved register into the stack frame.
       */
      FOREACH_BBL_INS(ibbl, iins) {
	if (REGSET_IS_MEMBER(AINS_REG_A(iins), regs_to_consider)
	    && AINS_CODE(iins) == I_stq
	    && AINS_REG_B(iins) == REG_SP
	    && AINS_DATA(iins) == SaveRegDispl(AINS_REG_A(iins))) {
	  break;
	}
      }
      /*
       * At this point, instructions that save callee-save registers into
       * the stack frame are killed, while other instructions are moved
       * into the next block.
       */
      for ( ; VALID(iins); iins = AINS_NEXT(iins)) {
	if (REGSET_IS_MEMBER(AINS_REG_A(iins), regs_to_consider)
	    && AINS_CODE(iins) == I_stq
	    && AINS_REG_B(iins) == REG_SP
	    && AINS_DATA(iins) == SaveRegDispl(AINS_REG_A(iins))) {
	  regs_to_consider = REGSET_DIFF(regs_to_consider, REG2BIT(AINS_REG_A(iins)));
	  AinsKill(iins);
	  nins++;

	  if (xverbose) {
	    printf("@@@FUNC   %s   ins   1\n", AFUN_NAME(ifun));
	  }
	}
	else {
	  AinsMove(iins, jbbl);
	}
      }

      AbblMoveSuccEdges(ibbl, jbbl);

      iins = AinsNew();
      AinsMakeBsr(iins, r0);
      AINS_BBL(iins) = ibbl;
      AinsAppend(iins);
      ains++;
      
      ABBL_TYPE(ibbl) = BT_CALL;
      
      sbbl = (r > 0 ? saveregbbls[r-8] : saveregbbls[0]);

      AedgLinkFancy(AedgNewFancy(ibbl, jbbl, ET_FUN_LINK));
      iedg = AedgNewFancy(ibbl, sbbl, ET_FUN_CALL);
      AedgLinkFancy(iedg);

      for (i = (r > 0 ? r-8 : 0); i >= 0; i--) {
	ABBL_COUNT(saveregbbls[i]) += ABBL_COUNT(ibbl);
      }
    }
  } while (chg);

  InfoStatus(stdout,"\t[FactorSaveRegs]: savings %d (%d deleted, %d added)\n", nins-ains, nins, ains);

}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "RegsetHasHoles"

static BOOL RegsetHasHoles(REGSET rs)
{
  REG rmin, rmax, r;

  for (rmin = 0; rmin < 63; rmin++) {
    if (REGSET_IS_MEMBER(rmin, rs)) break;
  }

  rmax = MaxRegNo(rs);

  for (r = rmin; r <= rmax; r++) {
    if (!REGSET_IS_MEMBER(r, rs)) {
      return TRUE;
    }
  }

  return FALSE;
}

static REGSET FillRegSetHoles(REGSET rs, REGSET live)
{
  REG rmin, rmax, r;
  REGSET final = rs;

  for (rmin = 0; rmin < 63; rmin++) {
    if (REGSET_IS_MEMBER(rmin, rs)) break;
  }
  
  rmax = MaxRegNo(rs);

  for (r = rmin; r <= rmax; r++)
    if (!REGSET_IS_MEMBER(r, rs) && REGSET_IS_MEMBER(r,live))
      final = REGSET_ADD(r,final);
  
  return final;
}


/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "FactorRestoreRegs"

extern void FactorRestoreRegs()
{
  INDEX ifun, maxsavefun;
  INDEX ibbl, jbbl, sbbl;
  INDEX iedg, xedg;
  INDEX iins, jins;
  INDEX saveregbbls[8];
  int  i;
  REGSET regs, tmpregs, regs_to_consider;
  REG r;
  BOOL retreg_is_ra, updates_sp, contains_ret;
  NUMBER nins = 0, ains = 0;
  BOOL verbose = HasOption(PRINT_OPT_VERBOSE, "FactorRestoreRegs");
  BOOL xverbose = HasOption(PRINT_OPT_VERBOSE, "FuncNames");

  maxsavefun = INVALID;
  regs = REGS_NONE;
  FOREACH_LIVE_FUN(ifun) {
    ibbl = AFUN_BBL_FIRST(ifun);
    tmpregs = AbblSavesRegs(ibbl) & SAVE_REGS;
    if (!REGSET_IS_SUBSET(tmpregs, regs)) {
      regs = REGSET_UNION(regs, tmpregs);
      maxsavefun = ifun;
    }
  }

  sbbl = AbblNew();
  ABBL_FUN(sbbl) = maxsavefun;
  ABBL_TYPE(sbbl) = BT_RET;
  AbblAppendFancy(sbbl);
  
  iins = AinsNew();
  AinsMakeAddq(iins, REG_SP, REG_RA, REG_SP);  /* $sp := $26 + $sp */
  AINS_BBL(iins) = sbbl;
  AinsAppend(iins);
  ains++;
  
  iins = AinsNew();
  AinsMakeSubq(iins, REG_SP, REG_RA, REG_RA);  /* $26 := $sp - $26 */
  AINS_BBL(iins) = sbbl;
  AinsAppend(iins);
  ains++;
  
  iins = AinsNew();
  AinsMakeLdq(iins, REG_RA, REG_RA, 0);        /* $26 := 0($26) */
  AINS_BBL(iins) = sbbl;
  AinsAppend(iins);
  ains++;
  
  iins = AinsNew();
  AinsMakeRet(iins, REG_RA);                   /* ret ($26) */
  AINS_BBL(iins) = sbbl;
  AinsAppend(iins);
  ains++;
  
  saveregbbls[0] = sbbl;
  
  i = 1;
  REGSET_FOREACH_REG(SAVE_REGS, r) {
    sbbl = AbblNew();
    ABBL_FUN(sbbl) = maxsavefun;
    ABBL_TYPE(sbbl) = BT_NORM;
    AbblAppendFancy(sbbl);
    saveregbbls[i] = sbbl;
    AedgLinkFancy(AedgNewFancy(saveregbbls[i], saveregbbls[i-1], ET_NORM));
    i += 1;
    
    iins = AinsNew();
    AinsMakeLdq(iins, REG_SP, r, SaveRegDispl(r));
    AINS_BBL(iins) = sbbl;
    AinsAppend(iins);
    ains++;
  }
  
  FOREACH_LIVE_FUN(ifun) {
    if (AFUN_IS_PSEUDO(ifun)) continue;

    retreg_is_ra = updates_sp = contains_ret = FALSE;
    FOREACH_BBL_PRED(AFUN_BBL_LAST(ifun), xedg) {
      if (AEDG_IS_INTER(xedg)) continue;

      ibbl = AEDG_FROM(xedg);
      if (ABBL_TYPE(ibbl) != BT_RET) continue;
      if (AbblBadRestore(ibbl)) continue;
      if (ibbl == saveregbbls[0]) continue;
      
      FOREACH_BBL_INS_R(ibbl, iins) {
	if (AINS_CODE(iins) == I_ret) {
	  contains_ret = TRUE;
	  retreg_is_ra = (AINS_REG_B(iins) == REG_RA);
	}
	else if (AINS_CODE(iins) == I_lda
		 && AINS_REG_B(iins) == REG_SP 
		 && AINS_REG_C(iins) == REG_SP) {
	  updates_sp = TRUE;
	}
      }
      
      if ( !(contains_ret && retreg_is_ra && updates_sp) ) break;

      regs_to_consider = AbblRestoresRegs(ibbl);

      if (MinRegNo(ABBL_OUT(ibbl) & SAVE_REGS)<MinRegNo(regs_to_consider))
	continue;	

      FillRegSetHoles(regs_to_consider,SAVE_REGS); // UNTESTED
      
      while (RegsetHasHoles(regs_to_consider)) {
	r = MaxRegNo(regs_to_consider);
	regs_to_consider = REGSET_DIFF(regs_to_consider, REGSET_SINGLETON(r));
      }

      r = MaxRegNo(regs_to_consider);
      if (!VALID(r)) continue;

      if (verbose) {
	printf("\t@@@ FactorRestoreRegs: bbl %d fun %d max. reg: %d\n", ibbl,ABBL_FUN(ibbl),r);
      }

      jins = INVALID;
      FOREACH_BBL_INS_R(ibbl, iins) {
	if (REGSET_IS_MEMBER(AINS_REG_C(iins), (regs_to_consider | B_RA))
	    && AINS_CODE(iins) == I_ldq
	    && AINS_REG_B(iins) == REG_SP
	    && AINS_DATA(iins) == SaveRegDispl(AINS_REG_C(iins))) {
	  AinsKill(iins);
	  nins++;

	  if (xverbose) {
	    printf("@@@FUNC   %s   ins   1\n", AFUN_NAME(ifun));
	  }
	}
	else if (AINS_CODE(iins) == I_ret) {
	  AinsKill(iins);
	  nins++;

	  if (xverbose) {
	    printf("@@@FUNC   %s   ins   1\n", AFUN_NAME(ifun));
	  }
	}
	else if (AINS_CODE(iins) == I_lda 
		 && AINS_REG_B(iins) == REG_SP
		 && AINS_REG_C(iins) == REG_SP
		 && !VALID(jins)) {
	  jins = iins;
	  AinsKill(iins);
	  nins++;

	  if (xverbose) {
	    printf("@@@FUNC   %s   ins   1\n", AFUN_NAME(ifun));
	  }
	}
      }
      iins = AinsNew();
      AinsMakeLda(iins, REG_ZERO, REG_RA, AINS_DATA(jins));
      AINS_BBL(iins) = ibbl;
      AinsAppend(iins);
      ains++;
      
      ABBL_TYPE(ibbl) = BT_NORM;
      
      FOREACH_BBL_SUCC(ibbl, iedg) {
	jbbl = AEDG_TAIL(iedg);
	if (ABBL_TYPE(jbbl) == BT_EXIT) {
	  AedgKillFancy(iedg);
	  AedgLinkFancy(AedgNewFancy(saveregbbls[0], jbbl, ET_EXIT_LINK));

	  for (i = (r > 0 ? r-8 : 0); i >= 0; i--) {
	    ABBL_COUNT(saveregbbls[i]) += ABBL_COUNT(ibbl);
	  }
	}
      }
      
      sbbl = (r > 0 ? saveregbbls[r-8] : saveregbbls[0]);
      AedgLinkFancy(AedgNewFancy(ibbl, sbbl, ET_NORM));
    }
    if ( !(contains_ret && retreg_is_ra && updates_sp) ) {
      continue;
    }
  }

  InfoStatus(stdout, "\t[FactorRestoreRegs]: savings %d (%d deleted, %d added)\n", nins-ains, nins, ains);
}

/*=======================================================
@FUN
AbblBadRestore() checks the following: (1) that any callee-saved
register that is restored from the block ibbl is restored from
the "right" slot in the stack frame; (2) the return address register
($26) is restored from 0($sp); (3) if the stack pointer is
redefined, it is computed from the value of $sp; and (4) if any of
the callee-saved registers/return address register/stack pointer
are used after they are updated.
@END
=======================================================*/
#undef __PROC__
#define __PROC__ "AbblBadRestore"

extern BOOL AbblBadRestore(INDEX ibbl)
{
  INDEX iins, jins;
  BOOL ra_defined = FALSE;
  REG r;

  CASSERT((VALID(ibbl)), ("invalid argument: %d\n", ibbl));

  if (ABBL_TYPE(ibbl) == BT_EXIT) {
    return FALSE;
  }
  /*
   * find the last load of the register.
   */
  FOREACH_BBL_INS_R(ibbl, iins) {
    if (AINS_CODE(iins) == I_ldq) {
      r = AINS_REG_C(iins);
      if (REGSET_IS_MEMBER(r, SAVE_REGS)) {
	if (AINS_REG_B(iins) != REG_SP) 
	return TRUE;
	if (AINS_DATA(iins) != SaveRegDispl(r))
	return TRUE;
	for (jins = AINS_NEXT(iins); VALID(jins); jins = AINS_NEXT(jins)) {
	  if (REGSET_IS_MEMBER(r, AinsUse(jins))) {
	    return TRUE;
	  }
	}
      }
      if (r == REG_RA) {
	ra_defined = TRUE;
	if ( !(AINS_REG_B(iins) == REG_SP && AINS_DATA(iins) == 0) )
	return TRUE;
	for (jins = AINS_NEXT(iins); VALID(jins); jins = AINS_NEXT(jins)) {
	  if (AINS_CODE(jins) != I_ret && REGSET_IS_MEMBER(r, AinsUse(jins))) {
	    return TRUE;
	  }
	}
      }
    }
    else if (AINS_CODE(iins) == I_lda && AINS_REG_C(iins) == REG_SP) {
      if (AINS_REG_B(iins) != REG_SP) 
      return TRUE;
      for (jins = AINS_NEXT(iins); VALID(jins); jins = AINS_NEXT(jins)) {
	if (REGSET_IS_MEMBER(REG_SP, AinsUse(jins))) {
	  return TRUE;
	}
      }
    }
  }

  return !ra_defined;
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "AbblBadSave"

static BOOL AbblBadSave(INDEX ibbl)
{
  INDEX iins, jins;
  BOOL ra_saved = FALSE;
  REG r;
  REGSET regs_considered = SAVE_REGS;

  CASSERT((VALID(ibbl)), ("invalid argument: %d\n", ibbl));

  /*
   * find the first store of the register.
   */
  FOREACH_BBL_INS(ibbl, iins) {
    if (AINS_CODE(iins) == I_stq) {
      r = AINS_REG_A(iins);
      if (REGSET_IS_MEMBER(r, regs_considered)) {
	if (AINS_REG_B(iins) != REG_SP) {
	  return TRUE;
	}
	if (AINS_DATA(iins) != SaveRegDispl(r)) {
	  return TRUE;
	}
	for (jins = AINS_PREV(iins); VALID(jins); jins = AINS_PREV(jins)) {
	  if (REGSET_IS_MEMBER(r, AinsUse(jins))) {
	    return TRUE;
	  }
	}
      }
      if (r == REG_RA) {
	ra_saved = TRUE;
	if ( !(AINS_REG_B(iins) == REG_SP && AINS_DATA(iins) == 0) ) {
	  return TRUE;
	}
	for (jins = AINS_PREV(iins); VALID(jins); jins = AINS_PREV(jins)) {
	  if (REGSET_IS_MEMBER(r, AinsUse(jins))) {
	    return TRUE;
	  }
	}
      }
      regs_considered = REGSET_DIFF(regs_considered, REG2BIT(r));
    }
    else if (AINS_CODE(iins) == I_lda && AINS_REG_C(iins) == REG_SP) {
      if (AINS_REG_B(iins) != REG_SP) {
	return TRUE;
      }
    }
  }

  return !ra_saved;
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "NumRegs"

static NUMBER NumRegs(REGSET regs)
{
  NUMBER nregs = 0;
  REG r;

  REGSET_FOREACH_REG(regs, r) {
    nregs += 1;
  }

  return nregs;
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "MaxRegNo"

static REG MaxRegNo(REGSET regs)
{
  REG r, maxr;

  maxr = INVALID;
  REGSET_FOREACH_REG(regs, r) {
    if (r > maxr) maxr = r;
  }
  return maxr;
}
/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "MinRegNo"

static REG MinRegNo(REGSET regs)
{
  REG r, minr;

  minr = 64;
  REGSET_FOREACH_REG(regs, r) {
    if (r < minr) minr = r;
  }
  return minr==64?-1:minr;
}

/*=======================================================
@FUN
AbblRedefinesRegs(ibbl) computes the set of registers that are
defined in block ibbl, not counting as definitions the restoring
of registers from memory in any return block of a function.
@END
=======================================================*/
#undef __PROC__
#define __PROC__ "AbblRedefinesRegs"

extern REGSET AbblRedefinesRegs(INDEX ibbl)
{
  INDEX iins;
  REGSET iregs;
  REG r;

  CASSERT((VALID(ibbl)), ("invalid argument: %d\n", ibbl));

  if (ABBL_TYPE(ibbl) == BT_EXIT) {
    return REGS_NONE;
  }

  if (ABBL_TYPE(ibbl) != BT_RET) {
    return AbblRegsDef(ibbl);
  }
  /*
   * work back to the last load of the register; then see if
   * that register is redefined earlier in the block.
   */
  iregs = REGS_NONE;

  REGSET_FOREACH_REG(SAVE_REGS, r) {
    FOREACH_BBL_INS_R(ibbl, iins) {
      if ( (AINS_CODE(iins) == I_ldq || AINS_CODE(iins) == I_ldt)
	   && AINS_REG_C(iins) == r) {
	break;
      }
    }
    for (iins = AINS_PREV(iins); VALID(iins); iins = AINS_PREV(iins)) {
      if (AINS_REG_C(iins) == r) {
	iregs |= REG2BIT(r);
	break;
      }
    }
  }

  return iregs;
}
	  
/*=======================================================
@FUN
AbblRestoresRegs(ibbl) computes the set of registers that are
restored in block ibbl, i.e., loaded into a register that is
not subsequently modified in ibbl.
@END
=======================================================*/
#undef __PROC__
#define __PROC__ "AbblRestoresRegs"

extern REGSET AbblRestoresRegs(INDEX ibbl)
{
  INDEX iins;
  REGSET iregs;
  REG r;

  CASSERT((VALID(ibbl)), ("invalid argument: %d\n", ibbl));

  if (ABBL_TYPE(ibbl) == BT_EXIT) {
    return REGS_NONE;
  }
  /*
   * find the last load of the register.
   */
  iregs = REGS_NONE;
  REGSET_FOREACH_REG(SAVE_REGS, r) {
    FOREACH_BBL_INS_R(ibbl, iins) {
      if ( (AINS_CODE(iins) == I_ldq || AINS_CODE(iins) == I_ldt)
	   && AINS_REG_C(iins) == r) {
	iregs |= REG2BIT(r);
	break;
      }
    }
  }

  return iregs;
}

/*=======================================================
@FUN
AbblSavesRegs(ibbl) computes the set of registers that are
saved in block ibbl, i.e., stored into memory prior to any
modification.
@END
=======================================================*/
#undef __PROC__
#define __PROC__ "AbblSavesRegs"

extern REGSET AbblSavesRegs(INDEX ibbl)
{
  INDEX iins;
  REGSET iregs, candidates;
  REG r;

  CASSERT((VALID(ibbl)), ("invalid argument: %d\n", ibbl));

  if (ABBL_TYPE(ibbl) == BT_EXIT) {
    return REGS_NONE;
  }
  /*
   * find the first store of the register.
   */
  candidates = (SAVE_REGS | FSAVE_REGS | B_RA);
  iregs = REGS_NONE;

  FOREACH_BBL_INS(ibbl, iins) {
    r = AINS_REG_A(iins);
    if ( (AINS_CODE(iins) == I_stq || AINS_CODE(iins) == I_stt)
	 && AINS_REG_B(iins) == REG_SP
	 && REGSET_IS_MEMBER(r, candidates) ) {
      iregs |= REG2BIT(r);
    }
    candidates = REGSET_DIFF(candidates, REG2BIT(AINS_REG_C(iins)));
  }

  return iregs;
}

#undef __PROC__
#define __PROC__ "AbblSavesRA"

static BOOL AbblSavesRA(INDEX ibbl)
{
  INDEX iins;

  CASSERT((VALID(ibbl)), ("invalid argument: %d\n", ibbl));

  /*
   * find the first store of the register.
   */
  FOREACH_BBL_INS(ibbl, iins) {
    if ( (AINS_CODE(iins) == I_stq)
	 && AINS_REG_A(iins) == REG_RA
	 && AINS_REG_B(iins) == REG_SP) {
      return TRUE;
    }
  }

  return FALSE;
}

/*=======================================================
@FUN
AbblUsesRegs(ibbl) computes the set of registers that are
used in block ibbl, not counting as uses the saving of registers
to memory in the initial block of a function.
@END
=======================================================*/
#undef __PROC__
#define __PROC__ "AbblUsesRegs"

extern REGSET AbblUsesRegs(INDEX ibbl)
{
  INDEX iins;
  REGSET iregs;
  REG r;

  CASSERT((VALID(ibbl)), ("invalid argument: %d\n", ibbl));

  if (ABBL_TYPE(ibbl) == BT_EXIT) {
    return REGS_NONE;
  }

  iregs = REGS_NONE;
  if (ibbl != AFUN_BBL_FIRST(ABBL_FUN(ibbl))) {
    FOREACH_BBL_INS(ibbl, iins) {
      iregs |= AinsUse(iins);
    }
    return iregs;
  }
  else {
    REGSET_FOREACH_REG(SAVE_REGS, r) {
      FOREACH_BBL_INS(ibbl, iins) {
	if ( (AINS_CODE(iins) == I_stq || AINS_CODE(iins) == I_stt)
	     && AINS_REG_A(iins) == r) {
	  break;
	}
      }
      for (iins = AINS_NEXT(iins); VALID(iins); iins = AINS_NEXT(iins)) {
	if (AINS_REG_A(iins) == r) {
	  iregs |= REG2BIT(r);
	  break;
	}
      }
    }
  }

  return iregs;
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "AinsCommutative"

extern BOOL AinsCommutative(INDEX iins)
{
  switch (AINS_CODE(iins)) {
  case I_addl:
  case I_addq:
  case I_cmpeq:
  case I_and:
  case I_bis:
  case I_cmoveq:
  case I_cmovne:
  case I_xor:
  case I_eqv:
  case I_mull:
  case I_mulq:
  case I_umulh:
  case I_addf:
  case I_addg:
  case I_adds:
  case I_addt:
  case I_mulf:
  case I_mulg:
    return TRUE;
  default:
    return FALSE;
  }
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "AinsIdentical"

extern BOOL AinsIdentical(INDEX i1, INDEX i2)
{
  if ( !(VALID(i1) && VALID(i2)) ) 
    return FALSE;

  if (AINS_CODE(i1) == I_bsr && AINS_CODE(i2) == I_bsr) {
    INDEX target1 = AEDG_TAIL(ABBL_EDG_CALL(AINS_BBL(i1)));
    INDEX target2 = AEDG_TAIL(ABBL_EDG_CALL(AINS_BBL(i2)));

    if (target1 == target2 && AINS_REG_C(i1) == AINS_REG_C(i2)) {
      return TRUE;
    }
    else {
      return FALSE;
    }
  }

  if ( AINS_CODE(i1) == AINS_CODE(i2) 
       && AINS_DATA(i1) == AINS_DATA(i2)
       && AINS_REG_A(i1) == AINS_REG_A(i2) 
       && AINS_REG_B(i1) == AINS_REG_B(i2) 
       && AINS_REG_C(i1) == AINS_REG_C(i2) ) {
    return TRUE;
  }
  
  return FALSE;
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "AbblAccessesStack"

extern BOOL AbblAccessesStack(INDEX ibbl)
{
  INDEX iins;

  CASSERT(VALID(ibbl), ("bad argument\n"));

  FOREACH_BBL_INS(ibbl, iins) {
    if (AINS_IS_LOAD(iins) && AINS_REG_B(iins) == REG_SP) return TRUE;
    if (AINS_IS_STORE(iins) && AINS_REG_B(iins) == REG_SP) return TRUE;
  }
  return FALSE;
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "AbblNotRenamable"

extern BOOL AbblNotRenamable(INDEX bbl1, INDEX bbl2)
{
  INDEX i1, i2;

  for (i1 = ABBL_INS_FIRST(bbl1), i2 = ABBL_INS_FIRST(bbl2);
       VALID(i1) && VALID(i2);
       i1 = AINS_NEXT(i1), i2 = AINS_NEXT(i2)) {
    if (AINS_CODE(i1) != AINS_CODE(i2)) return TRUE;
    if (AINS_DATA(i1) != AINS_DATA(i2)) return TRUE;
  }

  return FALSE;
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "AbblIdentical"

extern BOOL AbblIdentical(INDEX bbl1, INDEX bbl2)
{
  INDEX i1, i2;

  for (i1 = ABBL_INS_FIRST(bbl1), i2 = ABBL_INS_FIRST(bbl2);
       VALID(i1) && VALID(i2);
       i1 = AINS_NEXT(i1), i2 = AINS_NEXT(i2)) {
    if (!AinsIdentical(i1, i2)) break;
  }

  if (VALID(i1) || VALID(i2)) return FALSE;

  return TRUE;
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "AbblCommonSuffixLength"

extern NUMBER AbblCommonSuffixLength(INDEX ibbl, INDEX jbbl)
{
  INDEX iins, jins;
  NUMBER n = 0;

  CASSERT((VALID(ibbl) && VALID(jbbl)), ("bad argument\n"));

  for (iins = ABBL_INS_LAST(ibbl), jins = ABBL_INS_LAST(jbbl);
       VALID(iins) && VALID(jins);
       iins = AINS_PREV(iins), jins = AINS_PREV(jins)) {
    if (!AinsIdentical(iins, jins)) {
      return n;
    }
    n++;
  }

  return n;
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "AbblSubseq"

extern BOOL AbblSubseq(INDEX bbl1, INDEX bbl2)
{
  INDEX i1, i2;
  INDEX ibbl1, ibbl2;
  NUMBER n1, n2;

  n1 = AbblNInss(bbl1);
  n2 = AbblNInss(bbl2);

  if (n1 == 0 || n2 == 0) return FALSE;

  if (n1 > n2) {
    ibbl1 = bbl1;
    ibbl2 = bbl2;
  }
  else {
    ibbl1 = bbl2;
    ibbl2 = bbl1;
  }

  i1 = ABBL_INS_FIRST(ibbl1);
  i2 = ABBL_INS_FIRST(ibbl2);
  while (TRUE) {
    if (!VALID(i1) && !VALID(i2)) {
      /*
	printf("@@@ IDENTICAL bbls [%d instrs]: %d and %d\n", n1, bbl1, bbl2);
      */
      return TRUE;
    }
    if (VALID(i1) && !VALID(i2)) {
      printf("@@@ SUBSEQ: %d [%d instrs] is a subseq of %d [%d instrs]\n",
	     ibbl2, AbblNInss(ibbl2), ibbl1, AbblNInss(ibbl1));
      return TRUE;
    }
    if (!VALID(i1) && VALID(i2)) {
      return FALSE;
    }

    if (AinsIdentical(i1, i2)) {
      i1 = AINS_NEXT(i1);
      i2 = AINS_NEXT(i2);
      continue;
    }
    else {
      i1 = AINS_NEXT(i1);
      i2 = ABBL_INS_FIRST(ibbl2);
      continue;
    }
  }

  return FALSE;
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "ComBBlNInss"

extern void ComBblNInss()
{
  INDEX ibbl;

  FOREACH_LIVE_BBL(ibbl) {
    ABBL_NINS(ibbl) = AbblNInss(ibbl);
  }
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "AbblCode"

extern UINT64 AbblCode(INDEX ibbl)
{
  int i;
  INDEX iins;
  UINT64 code;

  code = 0;
  i = 1;
  FOREACH_BBL_INS(ibbl, iins) {
    if (i > NUM_INSTS_ENCODED) 
      break;
    i += 1;
    code = (code << 8) | (OpcodeCode(AINS_CODE(iins)) & 0xff);
  }

  return code;
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "NumOneBits"

extern NUMBER NumOneBits(UINT64 x)
{
  INT32 n = 0;

  while (x != 0) {
    if ( (x&0x1) != 0 ) {
      n++;
    }
    x = x >> 1;
  }

  return n;
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "InitOpcodeHash"

static void InitOpcodeHash()
{
  int i, hashval;
  long bitmask, pow2, n;
  INDEX ibbl;
  INDEX iins;
  UINT8 code;
  UINT16 byte0, byte1;

  for (i=0; i<NUM_OPCODES; i++) {
    OpcodeCounts(i) = 0;
    OpcodeCode(i) = 0;
    Opcode(i) = i;
  }

  for (i=0; i < NUM_BUCKETS; i++) {
    HashTbl(i) = INVALID;
  }

  FOREACH_LIVE_INS(iins) {
    OpcodeCounts(AINS_CODE(iins)) += 1;
  }

  qsort(opcode_list, NUM_OPCODES, sizeof(NUMBER), cmp);

  code = 0;
  for (i = 0; i < NUM_TOP_OPCODES; i++) {
    code += 1;
    OpcodeCode( Opcode(i) ) = code;
  }

  bitmask = NUM_BUCKETS-1;

  for (n = NUM_BUCKETS, pow2 = 0; n > 1; n /= 2) {
    pow2++;
  }

  FOREACH_LIVE_BBL_R(ibbl) {
    /*
     * it's never worthwhile factoring any block containing 1 or fewer
     * instruction, which means these blocks aren't worth adding to the
     * hash table either.
     */
    if (!(ABBL_NINS(ibbl) > 1)) continue;

    ABBL_CODE(ibbl) = AbblCode(ibbl);
    byte0 = ABBL_CODE(ibbl) & bitmask;
    byte1 = (ABBL_CODE(ibbl) & (bitmask << pow2)) >> pow2;
    hashval = (byte0 ^ byte1) & bitmask;
    ABBL_HASH_LINK(ibbl) = HashTbl(hashval);
    HashTbl(hashval) = ibbl;
  }
#if 0
  {
    NUMBER n, hash_bucket;
    NUMBER zero, max, maxi, tot;
    INDEX i;

    printf("@@@ pow2 = %d, bitmask = %x\n", pow2, bitmask);

    max = zero = maxi = tot = 0;

    printf("@@@ HASH TABLE DISTRIBUTION:\n@HASH:   ");
    for (hash_bucket = 0; hash_bucket < NUM_BUCKETS; hash_bucket += 1) {
      n = 0;
      for (i = HashTbl(hash_bucket); VALID(i); i = ABBL_HASH_LINK(i)) {
	tot++;
	n++;
      }
      printf("%d; ", n);
      if (n == 0) zero++;
      if (n > max) { max = n; maxi = hash_bucket; }
      if (hash_bucket > 0 && hash_bucket % 20 == 0) printf("\n@HASH:   ");
    }
    printf("\n@@@ HASH: Zero = %d [out of %d], Max = %d, ave = %d\n", 
	   zero, NUM_BUCKETS, max, tot/NUM_BUCKETS);
    printf("--- ");
#if 0
    for (i = HashTbl(maxi); VALID(i); i = ABBL_HASH_LINK(i)) {
      printf("%lx  ", ABBL_CODE(i));
      if (i > 0 && i % 5 == 0) printf("\n--- ");
    }
#endif

    printf("@HASH: ------------------------------------------------------------\n");
  }
#endif

}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "InitFactor"

extern void InitFactor()
{
  INDEX ibbl;

  UNMARK_ALL_BBLS;

  FOREACH_LIVE_BBL(ibbl) {
    ABBL_MASTER(ibbl) = INVALID;
    //    ABBL_UNMARK(ibbl);
    ABBL_CODE(ibbl) = 0;
  }

  InitOpcodeHash();
}

/*=======================================================
Create a new bbl and insert it before ibbl.  Redirect
incoming edges to ibbl to be directed at this new bbl
instead, and add a new edge from this bbl to ibbl.
=======================================================*/
#undef __PROC__
#define __PROC__ "InsertBblBeforeFancy"

extern INDEX InsertBblBeforeFancy(INDEX ibbl)
{
  INDEX xbbl;
  BOOL verbose = HasOption(PRINT_OPT_VERBOSE, "InsertBbl");

  xbbl = AbblNew();
  ABBL_FUN(xbbl) = ABBL_FUN(ibbl);
  ABBL_TYPE(xbbl) = BT_NORM;
  ABBL_COUNT(xbbl) = ABBL_COUNT(ibbl);
  ABBL_OUT(xbbl) = AbblRegsLiveIn(ibbl);

  if (ABBL_IS_FIRST(ibbl)) {
    AbblAppendFirst(xbbl);
  }
  else {
    AbblAppendFancy(xbbl);
  }

  AbblMigrateRelocInfo(ibbl, xbbl);

  AbblMovePredEdges(ibbl, xbbl);

  AedgLinkFancy(AedgNewFancy(xbbl, ibbl, ET_NORM));

  UpdateDominatorInfo(ABBL_FUN(ibbl));

  if (verbose) {
    printf("\t*** InsertBblBefore: ibbl %d [type %c], new bbl %d [type %c]\n",
	   ibbl, ABBL_TYPE(ibbl), xbbl, ABBL_TYPE(xbbl));
  }
    
  return xbbl;
}

/*=======================================================
Create a new bbl and insert it after ibbl.  Redirect
outgoing edges from ibbl to be directed from this new bbl
instead, and add a new edge from ibbl to this bbl.
After this, the new bbl gets the type of ibbl, while
ibbl becomes an "ordinary" bbl.
=======================================================*/
#undef __PROC__
#define __PROC__ "InsertBblAfterFancy"

extern INDEX InsertBblAfterFancy(INDEX ibbl)
{
  INDEX jbbl;
  BOOL verbose = HasOption(PRINT_OPT_VERBOSE, "InsertBbl");

  jbbl = AbblNew();
  ABBL_TYPE(jbbl) = ABBL_TYPE(ibbl);
  ABBL_TYPE(ibbl) = BT_NORM;
  ABBL_FUN(jbbl) = ABBL_FUN(ibbl);
  AbblAppendFancy(jbbl);
  
  ABBL_COUNT(jbbl) = ABBL_COUNT(ibbl);
  
  if( ABBL_FLAGS(ibbl) & F_GPREL32 ) {
    ABBL_FLAGS(jbbl) |= F_GPREL32;
    AINS_FLAGS(ABBL_INS_FIRST(jbbl)) |= F_GPREL32;
  }

  ABBL_OUT(jbbl) = ABBL_OUT(ibbl);

  AbblMoveSuccEdges(ibbl, jbbl);
  AedgLinkFancy(AedgNewFancy(ibbl, jbbl, ET_NORM));

  UpdateDominatorInfo(ABBL_FUN(ibbl));

  if (verbose) {
    printf("\t*** InsertBblBefore: ibbl %d [type %c], new bbl %d [type %c]\n",
	   ibbl, ABBL_TYPE(ibbl), jbbl, ABBL_TYPE(jbbl));
  }
    
  return jbbl;
}

/*=======================================================
@FUN
Separates each control flow instruction into a separate basic 
block of its own.  This may allow the non-control-flow portion
to be factored.
@END
=======================================================*/
#undef __PROC__
#define __PROC__ "SeparateControlFlowInstrs"

extern void SeparateControlFlowInstrs()
{
  INDEX ibbl, jbbl;
  INDEX iins, jins;

  FOREACH_LIVE_BBL_R(ibbl) {
    switch (ABBL_TYPE(ibbl)) {
    case BT_BRANCH:
    case BT_JUMP:
    case BT_SWITCH:
      iins = ABBL_INS_LAST(ibbl);
      if (VALID(AINS_PREV(iins))) {
        jbbl = InsertBblAfterFancy(ibbl);
	AinsMove(iins, jbbl);
      }
      break;

    case BT_CALL:
      if (ABBL_IS_FIRST(ibbl)) continue;

      iins = ABBL_INS_LAST(ibbl);
      if (VALID(AINS_PREV(iins))) {
        jbbl = InsertBblBeforeFancy(ibbl);
	FOREACH_BBL_INS(ibbl, jins) {
	  if (jins == iins) break;
	  AinsMove(jins, jbbl);
	}
      }
      break;
    }
  }
}

/*=======================================================
Rename registers in jbbl, if possible, so that jbbl becomes
identical to ibbl.
=======================================================*/
#undef __PROC__
#define __PROC__ "RenameBbl"

extern BOOL RenameBbl(INDEX ibbl, INDEX jbbl)
{
  INDEX iins, jins;
  BOOL verbose = HasOption(PRINT_OPT_VERBOSE, "FactorRenameBbl");
  
  //  if (jbbl<500 || jbbl>1000) return FALSE;

  if (AbblNotRenamable(ibbl, jbbl)) 
    return FALSE;
  if (ABBL_TYPE(ibbl) == BT_RET || ABBL_TYPE(ibbl) == BT_CALL) 
    return FALSE;
  if (ABBL_TYPE(jbbl) == BT_RET || ABBL_TYPE(jbbl) == BT_CALL) 
    return FALSE;
  if (ABBL_NINS(jbbl) <= 1) 
    return FALSE;

  /*****************************************
   start of new rename calculations
  *****************************************/
  if (CompareDagBbls(ibbl,jbbl))
    if (DetermineSubsts(ibbl,jbbl))
      {

	/*	if (jbbl<=1000)
	{
	  printf("rename bbl %d to %d\n",jbbl,ibbl);
	  PrintBblSuper(stdout,ibbl);
	  PrintRegSet(ABBL_OUT(ibbl));printf("\n");
	  PrintBblSuper(stdout,jbbl);
	  PrintRegSet(ABBL_OUT(jbbl));printf("\n");
	  }*/
	
	for (jins = ABBL_INS_LAST(jbbl),iins = ABBL_INS_LAST(ibbl);
	     VALID(jins) && VALID(iins);
	     jins=AINS_PREV(jins),iins=AINS_PREV(iins))	
	  {
	    AINS_REG_A(jins)=AINS_REG_A(iins);
	    AINS_REG_B(jins)=AINS_REG_B(iins);
	    AINS_REG_C(jins)=AINS_REG_C(iins);
	  }
	return TRUE;
      }
  
  return FALSE;
  /*****************************************
   end of new rename calculations
  *****************************************/
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "AbblRenameRegs"

extern BOOL AbblRenameRegs(INDEX ibbl, INDEX jbbl)
{
  INDEX table_commut[32];
  INT32 nr_commut = 0;
  INT32 i,j;
  INDEX iins;

  if (VALID(ABBL_MASTER(jbbl))) return FALSE;
  if (ABBL_IS_RENAMED(jbbl)) return FALSE;   /* already renamed */
  /*  
      if (AbblNInss(jbbl)<=32)
      {
      FOREACH_BBL_INS(ibbl,iins)
      {
      if (AinsCommutative(iins))
      table_commut[nr_commut++]=iins;
      }
      }
      
      nr_commut=0;
      
      for (i=0;i<(1<<nr_commut);i++)
      {
      for (j=0;j<nr_commut;j++)
      {
      if (i & 1<<j)
      {
      REG tmp = AINS_REG_A(table_commut[j]);
      AINS_REG_A(table_commut[j])=AINS_REG_B(table_commut[j]);
      AINS_REG_B(table_commut[j])=tmp;
      }
      }*/
  if (RenameBbl(ibbl, jbbl)) 
    {
      //	  if (i)
      //	    printf("extra one found\n");
      AddMoves(jbbl);
      return TRUE;
    }

  /*      
	  for (j=0;j<nr_commut;j++)
	  {
	  if (i & 1<<j)
	  {
	  REG tmp = AINS_REG_A(table_commut[j]);
	  AINS_REG_A(table_commut[j])=AINS_REG_B(table_commut[j]);
	  AINS_REG_B(table_commut[j])=tmp;
	  }
	  }
	  }*/
  return FALSE;
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "GroupIdenticalBbls"

extern bbl_lst_lst_ptr GroupIdenticalBbls(BOOL ret_only)
{
  INDEX ibbl, jbbl;
  INDEX m0, m1;
  int hash_bucket;
  bbl_lst_ptr bltmp;
  bbl_lst_lst_ptr bllroot, blltmp;

  for (hash_bucket = 0; hash_bucket < NUM_BUCKETS; hash_bucket += 1) {
    for (ibbl = HashTbl(hash_bucket); VALID(ibbl); ibbl = ABBL_HASH_LINK(ibbl)) {
      if (VALID(ABBL_MASTER(ibbl))) continue;
      if (AFUN_IS_PSEUDO(ABBL_FUN(ibbl))) continue;
      if (ret_only) {
	if (ABBL_TYPE(ibbl) != BT_RET) continue;
      }
      else if (ABBL_TYPE(ibbl) == BT_RET) continue;
      
      for (jbbl = ABBL_HASH_LINK(ibbl); VALID(jbbl); jbbl = ABBL_HASH_LINK(jbbl)) {
	if (ABBL_CODE(ibbl) != ABBL_CODE(jbbl)) continue;
	if (ABBL_TYPE(ibbl) != ABBL_TYPE(jbbl)) continue;
	if (ABBL_NINS(ibbl) != ABBL_NINS(jbbl)) continue;
	if (AFUN_IS_PSEUDO(ABBL_FUN(jbbl))) continue;
	if (VALID(ABBL_MASTER(jbbl))) continue;
	
	if (AbblIdentical(ibbl, jbbl) || AbblRenameRegs(ibbl, jbbl)) {
	  m0 = VALID(ABBL_MASTER(ibbl)) ? ABBL_MASTER(ibbl) : ibbl;
	  m1 = VALID(ABBL_MASTER(jbbl)) ? ABBL_MASTER(jbbl) : jbbl;

	  ABBL_MASTER(ibbl) = ABBL_MASTER(jbbl) = MIN(m0, m1);
	}
      }
    }
  }

  bllroot = NULL;

  FOREACH_LIVE_BBL(ibbl) {
    if (!VALID(ABBL_MASTER(ibbl))) continue;

    bltmp = (bbl_lst_ptr)AllocNode(sizeof(*bltmp));
    FACTOR_BBL(bltmp) = ibbl;
    Next(bltmp) = NULL;
    PROCESSED(bltmp) = FALSE;

    if (ABBL_MASTER(ibbl) == ibbl) { /* start new list */
      blltmp = (bbl_lst_lst_ptr)AllocNode(sizeof(*blltmp));
      FACTOR_ROOT(blltmp) = ibbl;
      FACTOR_BBL_LST(blltmp) = bltmp;
      PROCESSED(blltmp) = FALSE;
      Next(blltmp) = bllroot;
      bllroot = blltmp;
    }
    else { /* add to existing list */
      for (blltmp = bllroot; blltmp; blltmp = Next(blltmp)) {
	if (ABBL_MASTER(ibbl) == FACTOR_ROOT(blltmp)) {
	  break;
	}
      }
      CASSERT((blltmp != NULL), ("could not find entry: bbl = %d, master = %d\n",ibbl,ABBL_MASTER(ibbl)));
      Next(bltmp) = FACTOR_BBL_LST(blltmp);
      FACTOR_BBL_LST(blltmp) = bltmp;
    }
  }

  return bllroot;
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "FindRetReg"

extern REG FindRetReg(REGSET regs)
{
  REG r;

  CASSERT(regs != REGS_NONE, ("empty register set!\n"));

  if (REGSET_IS_MEMBER(REG_RA, regs)) {
    return REG_RA;
  }

  for (r = 28; r >= 0; r--) {
    if REGSET_IS_MEMBER(r, regs) {
      return r;
    }
  }

  return INVALID;
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "FactorRetBbls"

extern void FactorRetBbls()
{
  INDEX ibbl, jbbl, obbl, xbbl;
  INDEX iedg;
  NUMBER nbbls, savings;
  bbl_lst_lst_ptr bllroot, blltmp;
  bbl_lst_ptr blst;
  BOOL verbose = HasOption(PRINT_OPT_VERBOSE, "FactorRetBbl");
  BOOL xverbose = HasOption(PRINT_OPT_VERBOSE, "FuncNames");

  InitFactor();

  bllroot = GroupIdenticalBbls(TRUE);
  nbbls = savings = 0;

  for (blltmp = bllroot; blltmp != NULL; blltmp = Next(blltmp)) {
    if (ABBL_TYPE(FACTOR_ROOT(blltmp)) == BT_RET) {
      obbl = FACTOR_ROOT(blltmp);
      for (blst = FACTOR_BBL_LST(blltmp); blst; blst = Next(blst)) {
	ibbl = FACTOR_BBL(blst);

	if (xverbose) {
	  printf("@@@FUNC %s   ins  %d\n", AFUN_NAME(ABBL_FUN(ibbl)),
		 ABBL_NINS(ibbl));
	}

	if (ibbl == obbl) continue;

	/*
	 * alto/squeeze doesn't like switch edges that cross function 
	 * boundaries.  If any of the predecessors of ibbl is a switch
	 * node, insert a dummy node between that node and ibbl.
	 */
	FOREACH_BBL_PRED(ibbl, iedg) {
	  jbbl = AEDG_FROM(iedg);
	  if (ABBL_TYPE(jbbl) != BT_SWITCH) continue;

	  xbbl = AbblNew();
	  ABBL_FUN(xbbl) = ABBL_FUN(ibbl);
	  ABBL_COUNT(xbbl) = ABBL_COUNT(ibbl);
	  ABBL_TYPE(xbbl) = BT_NORM;
	  AbblAppendFancy(xbbl);

	  AbblMigrateRelocInfo(ibbl, xbbl);
	  AbblMovePredEdges(ibbl, xbbl);

	  AedgLinkFancy(AedgNewFancy(xbbl, ibbl, ET_NORM));
	}

	if (ABBL_IS_FIRST(ibbl))
	  {
	    xbbl = AbblNew();
	    ABBL_FUN(xbbl) = ABBL_FUN(ibbl);
	    ABBL_COUNT(xbbl) = ABBL_COUNT(ibbl);
	    ABBL_TYPE(xbbl) = BT_NORM;
	    AbblAppendFirst(xbbl);

	    AbblMigrateRelocInfo(ibbl, xbbl);
	    AbblMovePredEdges(ibbl, xbbl);
	    
	    AedgLinkFancy(AedgNewFancy(xbbl, ibbl, ET_NORM));
	  }

	AbblMigrateRelocInfo(ibbl, obbl);

	if (verbose) {
	  printf("@@@ FactorRetBbls: %d [ fun %d ] --> %d [ fun %d ] : Preds: ",
		 ibbl, ABBL_FUN(ibbl), obbl, ABBL_FUN(obbl));
	  FOREACH_BBL_PRED(ibbl, iedg) {
	    printf("%d [type %c, fun %d]; ", AEDG_FROM(iedg), 
		   ABBL_TYPE(AEDG_FROM(iedg)), ABBL_FUN(AEDG_FROM(iedg)));
	  }
	  printf("\n");
	}

	AbblMovePredEdges(ibbl, obbl);
	
	ABBL_COUNT(obbl) += ABBL_COUNT(ibbl);
	savings += ABBL_NINS(ibbl);
	AbblKill(ibbl);
	nbbls++;
      }
    }
  }
  InfoStatus(stdout, "\t[FactorRetBbls]: %d bbls, savings %d\n", nbbls, savings);
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "FactorBbls"

extern void FactorBbls()
{
  INDEX ifun;
  INDEX ibbl, obbl, rbbl;
  INDEX iedg;
  INDEX iins;
  NUMBER benefit[32], maxbenefit, addins, delins, curraddins, currdelins, savings;
  NUMBER nbbls_left = 0;
  bbl_lst_lst_ptr bllroot, blltmp;
  bbl_lst_ptr blst, bltmp;
  REGSET availregs;
  REG r, retreg;
  int i;
  BOOL chg;
  BOOL root_factored;
  BOOL verbose = HasOption(PRINT_OPT_VERBOSE, "FactorBbl");
  BOOL xverbose = HasOption(PRINT_OPT_VERBOSE, "FuncNames");

  bllroot = GroupIdenticalBbls(FALSE);
  ComLiveness(stdout);
  
  addins = delins = 0;

  for (blltmp = bllroot; blltmp != NULL; blltmp = Next(blltmp)) {
    if (PROCESSED(blltmp)) continue;
    if (ABBL_TYPE(FACTOR_ROOT(blltmp)) == BT_CALL ||
	ABBL_TYPE(FACTOR_ROOT(blltmp)) == BT_PAL  ||
	ABBL_TYPE(FACTOR_ROOT(blltmp)) == BT_RET  ||
	ABBL_TYPE(FACTOR_ROOT(blltmp)) == BT_BRANCH) {
      PROCESSED(blltmp) = TRUE;
      continue;
    }

    if (ABBL_TYPE(FACTOR_ROOT(blltmp)) == BT_RET) {
      obbl = FACTOR_ROOT(blltmp);
      for (blst = FACTOR_BBL_LST(blltmp); blst; blst = Next(blst)) {
	if (FACTOR_BBL(blst) == obbl) continue;

	ibbl = FACTOR_BBL(blst);

	FOREACH_BBL_PRED(ibbl, iedg) {
	  if (AEDG_TYPE(iedg) == ET_SWITCH || AEDG_TYPE(iedg) == ET_HELL) {
	    ABBL_FLAGS(obbl) |= F_RELOC;
	    AbblMigrateRelocInfo(ibbl, obbl);
	    break;
	  }
	}

	AbblMovePredEdges(ibbl, obbl);
	ABBL_COUNT(obbl) += ABBL_COUNT(ibbl);
	delins += ABBL_NINS(FACTOR_BBL(blst));
      }
      continue;
    }

    chg = TRUE;
    while (chg) {
      chg = FALSE;
      root_factored = FALSE;
      for (blst = FACTOR_BBL_LST(blltmp); blst; blst = Next(blst)) {
	if (FACTOR_BBL(blst) == FACTOR_ROOT(blltmp) && PROCESSED(blst)) {
	  root_factored = TRUE;
	  break;
	}
      }
      /*
       * if the root bbl has been factored, i.e., the instructions in it have
       * been replaced by a "bsr ...", we can't use it for any further factoring
       * of the blocks within this group.  In this case, if there are two or more
       * unprocessed blocks in this group, we choose the smallest numbered one
       * as the new root, redefine the ABBL_MASTER field of the unprocessed blocks
       * to be this one, and carry on.
       */
      if (root_factored) {
	nbbls_left = 0;
	obbl = INVALID;
	for (blst = FACTOR_BBL_LST(blltmp); blst; blst = Next(blst)) {
	  if (!PROCESSED(blst)) {
	    obbl = nbbls_left > 0 ? MIN(obbl, FACTOR_BBL(blst)) : FACTOR_BBL(blst);
	    nbbls_left++;
	  }
	}
	if (nbbls_left > 1) {
	  if (verbose) {
	    printf("@@@ Renaming root: %d --> %d\n", FACTOR_ROOT(blltmp), obbl);
	  }
	  FACTOR_ROOT(blltmp) = obbl;
	  for (blst = FACTOR_BBL_LST(blltmp); blst; blst = Next(blst)) {
	    if (!PROCESSED(blst)) {
	      ABBL_MASTER(FACTOR_BBL(blst)) = obbl;
	    }
	  }
	}
	else {
	  break;
	}
      }
    
      for (i = 0; i < 29; i++) benefit[i] = 0;

      for (blst = FACTOR_BBL_LST(blltmp); blst; blst = Next(blst)) {
	if (PROCESSED(blst)) continue;

	ibbl = FACTOR_BBL(blst);
	availregs = REGS_INT_BUT_ZERO & ~REG2BIT(REG_SP) & ~REG2BIT(REG_GP);
	availregs = REGSET_DIFF(availregs, SAVE_REGS);
	availregs = REGSET_DIFF(availregs, AbblRegsLiveIn(ibbl));
	availregs = REGSET_DIFF(availregs, AbblRegsDef(ibbl));

	REGSET_FOREACH_REG(availregs, r) {
	  benefit[r] += 1;
	}
      }

      maxbenefit = 0;
      for (i = 28; i >= 0; i--) {
	if (benefit[i] > maxbenefit) {
	  maxbenefit = benefit[i];
	  retreg = i;
	}
      }
      
      retreg = INVALID;
      if (maxbenefit == 0) {
	PROCESSED(blltmp) = TRUE;
	continue;
      }
      else {
	availregs = REGS_NONE;
	for (i=28; i >= 0; i--) {
	  if (benefit[i] == maxbenefit) {
	    availregs = REGSET_UNION(availregs, REG2BIT(i));
	  }
	}
	retreg = FindRetReg(availregs);
      }

      curraddins = ABBL_NINS(FACTOR_ROOT(blltmp))+maxbenefit+1;
      currdelins = ABBL_NINS(FACTOR_ROOT(blltmp))*maxbenefit;
      savings = currdelins-curraddins;

      if (savings <= 0) {
	PROCESSED(blltmp) = TRUE;
	continue;
      }
      /*
       * if necessary, create a new function with FACTOR_ROOT(blltmp) 
       * as the body and retreg as the return register.  Modify each bbl in the
       * group to call this new function.
       */
      if (verbose) {
	printf("@@@ FACTORING: Root %d [ fun %d ], bbls %d; addins %d, delins %d, savings %d\n", 
	       FACTOR_ROOT(blltmp), ABBL_FUN(FACTOR_ROOT(blltmp)),
	       maxbenefit,
	       curraddins, currdelins, currdelins-curraddins);
	printf("... [(%d instrs, R = %d, no. of bbls = %d)]: ", 
	       ABBL_NINS(FACTOR_ROOT(blltmp)), retreg, maxbenefit);
	for (bltmp = FACTOR_BBL_LST(blltmp); bltmp; bltmp = Next(bltmp)) {
	  if (PROCESSED(bltmp)) continue;
	  printf("%d ", FACTOR_BBL(bltmp));
	}
	printf("\n");
      }
      
      ifun = ibbl = INVALID;
      for (blst = FACTOR_BBL_LST(blltmp); blst; blst = Next(blst)) {
	if (PROCESSED(blst)) continue;
	if (REGSET_IS_MEMBER(retreg, AbblRegsLiveIn(FACTOR_BBL(blst)))) continue;

	if (!VALID(ifun)) {
	  /*
	   * create a function 
	   */
	  ifun = AfunNewFancy();
	  AFUN_NAME(ifun) = NewFunName("_sFactor", FACTOR_ROOT(blltmp));
	  
	  ibbl = AbblDup(FACTOR_ROOT(blltmp));
	  iins = AinsNew();
	  AinsMakeRet(iins, retreg);
	  AINS_BBL(iins) = ibbl;
	  AinsAppend(iins);
	  
	  ABBL_FUN(ibbl) = ifun;
	  AbblAppendFancy(ibbl);
	  ABBL_TYPE(ibbl) = BT_RET;
	  ABBL_NINS(ibbl) = ABBL_NINS(FACTOR_ROOT(blltmp));
	  
	  AedgLinkFancy( AedgNewFancy(ibbl, ibbl, ET_EXIT_LINK) );

	  if (AbblAccessesStack(ibbl)) {
	    AFUN_FLAGS(ifun) |= FF_STK_ARG;
	  }
	}

	if (AFUN_HAS_STK_ALIAS(ABBL_FUN(FACTOR_BBL(blst)))) {
	  AFUN_FLAGS(ifun) |= FF_STK_ALIAS;
	}
	if (AFUN_HAS_STK_INVALID(ABBL_FUN(FACTOR_BBL(blst)))) {
	  AFUN_FLAGS(ifun) |= FF_STK_INVALID;
	}
	AfunComputeStackAliases(ifun);

	AFUN_AUXFLAGS(ifun) |= FF_FACTORED;

	obbl = FACTOR_BBL(blst);

	//	if (ABBL_FUN(obbl)>=CONFIG_MIN && ABBL_FUN(obbl)<=CONFIG_MAX) continue;

	if (xverbose) {
	  printf("@@@FUNC %s   ins  %.2f\n", AFUN_NAME(ABBL_FUN(obbl)),
		 (double)savings/(double)maxbenefit);
	}

	PROCESSED(blst) = TRUE;
	FOREACH_BBL_INS(obbl, iins) {
	  AinsKill(iins);
	}
	iins = AinsNew();
	AinsMakeBsr(iins, retreg);
	AINS_BBL(iins) = obbl;
	AinsAppend(iins);
	
	rbbl = AbblNew();
	ABBL_FUN(rbbl) = ABBL_FUN(obbl);
	ABBL_COUNT(rbbl) = ABBL_COUNT(obbl);
	ABBL_TYPE(rbbl) = BT_NORM;
	AbblAppendFancy(rbbl);
	/*
	 * successors of obbl become successors of rbbl.
	 */
	FOREACH_BBL_SUCC(obbl, iedg) {
	  AedgLinkFancy(AedgNewFancy(rbbl, AEDG_TAIL(iedg), AEDG_TYPE(iedg)));
	  AedgKillFancy(iedg);
	}
	
	ABBL_TYPE(obbl) = BT_CALL;
	
	AedgLinkFancy(AedgNewFancy(obbl, rbbl, ET_FUN_LINK));
	AedgLinkFancy(AedgNewFancy(obbl, ibbl, ET_FUN_CALL));
	nbbls_left--;
	chg = TRUE;
      }
      ABBL_COUNT(AFUN_BBL_FIRST(ifun)) = AfunCallFreq(ifun);

      addins += curraddins;
      delins += currdelins;
    }
  }
  
  InfoStatus(stdout,"\t[FactorBbls]: savings %d (%d deleted, %d added)\n", delins-addins, delins, addins);
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "OptFactorSubgraphs"

extern void OptFactorSubgraphs()
{
  clock_t time = clock();

  if( HasOption(DISABLED_OPT,"Factor") ) return;

  InfoPhase(stdout, "Factoring");

  SeparateControlFlowInstrs();

  InfoProgress();

  ComBblNInss();

  InfoProgress();

  InitFactor();

  ComLiveness(stdout);

  InfoProgress();

  FactorBbls();

  ComLiveness(stdout);

  InfoProgress();

  /*************************************************************/

  if( HasOption(PRINT_OPT_WHEN,"Preregion") ) {
    PrintAll(stdout);
    exit(0);
  }

  ComLiveness(stdout);

  //  FactorRegions();

  if( HasOption(PRINT_OPT_WHEN,"Postregion") ) {
    PrintAll(stdout);
    exit(0);
  }

  InfoProgress();

  /*************************************************************/

  FactorRetBbls();

  InfoProgress();

  FactorSaveRegs();

  InfoProgress();

  FactorRestoreRegs();

  InfoProgress();

  time = (clock() - time) * 1000 / CLOCKS_PER_SEC;
  InfoStatus(stdout, "\t[OptFactorSubgraphs]: { %d ms }\n", time);
}

static BOOL CompareDagBbls(INDEX ibbl,INDEX jbbl)
{
  INDEX jins,iins,ji,ii;
  REG rja,rjb,rjc;
  REG ria,rib,ric;

  for (jins = ABBL_INS_FIRST(jbbl),iins = ABBL_INS_FIRST(ibbl);
       VALID(jins) && VALID(iins);
       jins=AINS_NEXT(jins),iins=AINS_NEXT(iins))
    {
      ria = AINS_REG_A(iins);
      rib = AINS_REG_B(iins);
      ric = AINS_REG_C(iins);
      
      rja = AINS_REG_A(jins);
      rjb = AINS_REG_B(jins);
      rjc = AINS_REG_C(jins);
      
      for (ii=AINS_PREV(iins),ji=AINS_PREV(jins);
	   VALID(ii) && VALID(ji);
	   ii=AINS_PREV(ii),ji=AINS_PREV(ji))
	{
	  if (AINS_REG_C(ii) == ria && AINS_REG_C(ji) != rja) return FALSE;
	  if (AINS_REG_C(ii) != ria && AINS_REG_C(ji) == rja) return FALSE;
	  if (AINS_REG_C(ii) == ria && AINS_REG_C(ji) == rja) ria=rja=INVALID; // defs have been found 
	  if (AINS_REG_C(ii) == rib && AINS_REG_C(ji) != rjb) return FALSE;
	  if (AINS_REG_C(ii) != rib && AINS_REG_C(ji) == rjb) return FALSE;
	  if (AINS_REG_C(ii) == rib && AINS_REG_C(ji) == rjb) rib=rjb=INVALID;// defs have been found
	  if (AINS_TYPE(ii) == IT_ICM || AINS_TYPE(ii) == IT_FCM)
	    {
	      if (AINS_REG_C(ii) == ric && AINS_REG_C(ji) != rjc) return FALSE;
	      if (AINS_REG_C(ii) != ric && AINS_REG_C(ji) == rjc) return FALSE;
	      if (AINS_REG_C(ii) == ric && AINS_REG_C(ji) == rjc) ric=rjc=INVALID;// defs have been found
	    }
	}
    }
  return TRUE;

}
#define table(i)  rotate_table[i]
#define get_table_bit(i,j) rotate_table[i]&(1L<<j)
#define set_table_bit(i,j) rotate_table[i]|=(1L<<j)

static BOOL DetermineSubsts(INDEX ibbl,INDEX jbbl)
{
  INDEX jins,iins;
  REG rja,rjb,rjc;
  REG ria,rib,ric;
  INT32 i,j;
  REGSET defined = REGS_NONE;
  REGSET ilive,jlive;
  REG reg;
  int n;
  BOOL change;
  REGSET temp_regs;
  REGSET regs_to_spill;
  BOOL flag = 0;

  for (i = 0; i < 64; i++) {
    InSubst(i) = OutSubst(i) = INVALID;
  }  

  for (jins = ABBL_INS_FIRST(jbbl),iins = ABBL_INS_FIRST(ibbl);
       VALID(jins) && VALID(iins);
       jins=AINS_NEXT(jins),iins=AINS_NEXT(iins))
    {
      ria = AINS_REG_A(iins);
      rib = AINS_REG_B(iins);
      ric = AINS_REG_C(iins);
      
      rja = AINS_REG_A(jins);
      rjb = AINS_REG_B(jins);
      rjc = AINS_REG_C(jins); 
      
      if (!REGSET_IS_MEMBER(ria,defined))
	{
	  if (VALID(InSubst(ria)) && rja!=InSubst(ria)) 
	    return FALSE;
	  InSubst(ria)=rja;
	  //	  printf("RA precopy %d -> %d\n",rja,ria);
	}
      if (!REGSET_IS_MEMBER(rib,defined))
	{
	  if (VALID(InSubst(rib)) && rjb!=InSubst(rib)) 
	    return FALSE;
	  InSubst(rib)=rjb;
	  //	  printf("RA precopy %d -> %d\n",rjb,rib);
	}
      if (AINS_TYPE(iins) == IT_ICM || AINS_TYPE(iins) == IT_FCM)
	{
	  if (!REGSET_IS_MEMBER(ric,defined))
	    {
	      if (VALID(InSubst(ric)) && rjc!=InSubst(ric)) 
		return FALSE;
	      InSubst(ric)=rjc;
	    }
	}

      defined = REGSET_ADD(ric,defined);    
    }
  
  defined = REGS_NONE;
  ilive = REGS_ALL;
  jlive = ABBL_OUT(jbbl);

  for (jins = ABBL_INS_LAST(jbbl),iins = ABBL_INS_LAST(ibbl);
       VALID(jins) && VALID(iins);
       jins=AINS_PREV(jins),iins=AINS_PREV(iins))
    {
      ria = AINS_REG_A(iins);
      rib = AINS_REG_B(iins);
      ric = AINS_REG_C(iins);
      
      rja = AINS_REG_A(jins);
      rjb = AINS_REG_B(jins);
      rjc = AINS_REG_C(jins); 
      
      if (REGSET_IS_MEMBER(rjc,jlive) && !REGSET_IS_MEMBER(ric,ilive))
	return FALSE;
      
      if (REGSET_IS_MEMBER(rjc,jlive))
	OutSubst(ric)=rjc;
      
      jlive = REGSET_DIFF(jlive,REGSET_SINGLETON(rjc));
      ilive = REGSET_DIFF(ilive,REGSET_SINGLETON(ric));
    }
  
  // live in - live out zonder def mag niet naar gerenamed doen : UITBREIDING: tijdelijk opslaan in ander register als dit de moeite loont
  
#if 1
  temp_regs = ~AbblRegsDef(ibbl) & ~AbblRegsUse(ibbl) & ~AbblRegsLiveIn(jbbl) & REGS_ALL_BUT_ZERO & ~REG2BIT(REG_SP) & ~REG2BIT(REG_GP);

  REGSET_FOREACH_REG(REGS_ALL,reg)
    {
      if (VALID(InSubst(reg)) && InSubst(reg)!=reg)
	{
	  temp_regs &= ~REG2BIT(reg);
	  temp_regs &= ~REG2BIT(InSubst(reg));
	}
      if (VALID(OutSubst(reg)) && OutSubst(reg)!=reg)
	{
	  temp_regs &= ~REG2BIT(reg);
	  temp_regs &= ~REG2BIT(OutSubst(reg));
	}
    }

  regs_to_spill = AbblRegsDef(ibbl) & AbblRegsLiveIn(jbbl) & ABBL_OUT(jbbl) & ~AbblRegsDef(jbbl) & REGS_ALL_BUT_ZERO;

  REGSET_FOREACH_REG(AbblRegsLiveIn(jbbl) & ABBL_OUT(jbbl) & ~AbblRegsDef(jbbl) & REGS_ALL_BUT_ZERO,reg)
    {
      if (VALID(InSubst(reg)) && InSubst(reg)!=reg)
	regs_to_spill |= REG2BIT(reg);
    }
  
  REGSET_FOREACH_REG(regs_to_spill,reg)
    {
      REG spill_reg = INVALID;
      spill_reg = (REG2BIT(reg)&REGS_INT) ? RegsGetFirstInt(temp_regs) : RegsGetFirstFlt(temp_regs);
      if (!VALID(spill_reg))
	return FALSE;
      InSubst(spill_reg)=reg;
      OutSubst(spill_reg)=reg;
      temp_regs &= ~REG2BIT(spill_reg);
    }
#else
  if (AbblRegsDef(ibbl) & AbblRegsLiveIn(jbbl) & ABBL_OUT(jbbl) & ~AbblRegsDef(jbbl) & REGS_ALL_BUT_ZERO)
    return FALSE;
  
  REGSET_FOREACH_REG(AbblRegsLiveIn(jbbl) & ABBL_OUT(jbbl) & ~AbblRegsDef(jbbl) & REGS_ALL_BUT_ZERO,reg)
    {
      if (VALID(InSubst(reg)) && InSubst(reg)!=reg)
	return FALSE;
    }
#endif
  // zero register mag niet naar hernoemd worden

  if (VALID(InSubst(REG_ZERO)) && InSubst(REG_ZERO)!=REG_ZERO)
    return FALSE;
  if (VALID(InSubst(REG_FZERO)) && InSubst(REG_FZERO)!=REG_FZERO)
    return FALSE;
  if (VALID(InSubst(REG_SP)) && InSubst(REG_SP)!=REG_SP)
    return FALSE;

  // tel aantal extra instructies

  n=0;

  for (reg = 0; reg < 64; reg++)
    {
      if (VALID(InSubst(reg)) && InSubst(reg)!=reg)
	n++;
      if (VALID(OutSubst(reg)) && OutSubst(reg)!=reg)
	n++;
    }
   
  if (n+1>=ABBL_NINS(ibbl))
    return FALSE;


  // remove nops
   
  for (i=0; i<64; i++)
    {
      if (VALID(InSubst(i)))
	if (InSubst(i)==i)
	  InSubst(i)=INVALID;
      if (VALID(OutSubst(i)))
	if (OutSubst(i)==i)
	  OutSubst(i)=INVALID;
    }
  
  // do a loop test: there may be no loops like r5->r7 r7->r4 r4->r5

  for (i=0;i<64;i++)
    table(i)=0L;

  for (i=0;i<64;i++)
    {
      if (VALID(InSubst(i)))
	{
	  //	  printf("bits to set %d -> %d\n",InSubst(i),i);
	  set_table_bit(InSubst(i),i);
	}
    }

  change = 1;

  while (change)
    {
      INT64 old;
      INT64 new;
      change = 0;
      for (i=0;i<64;i++)
	for (j=0;j<64;j++)
	  {
	    if (get_table_bit(i,j))
	      {
		old = table(i);
		new = table(i)|table(j);
		if (new!=old)
		  {
		    table(i)=new;
		    change = 1;
		  }
	      }
	  }
    }

  for (i=0;i<64;i++)
    {
      if (get_table_bit(i,i))
	return FALSE;
    }
   
  for (i=0;i<64;i++)
    table(i)=0L;

  for (i=0;i<64;i++)
    {
      if (VALID(OutSubst(i)))
	{
	  set_table_bit(OutSubst(i),i);
	}
    }

  change = 1;

   
  while (change)
    {
      INT64 old;
      INT64 new;
      change = 0;
      for (i=0;i<64;i++)
	for (j=0;j<64;j++)
	  {
	    if (get_table_bit(i,j))
	      {
		old = table(i);
		new = table(i)|table(j);
		if (new!=old)
		  {
		    table(i)=new;
		    change = 1;
		  }
	      }
	  }
    }
 
  for (i=0;i<64;i++)
    if (get_table_bit(i,i))
      return FALSE;
   
  if (n > 0) {
    ABBL_AUXFLAGS(jbbl) |= FB_RENAMED;
  }

  return TRUE;

}

static void AddMoves(INDEX jbbl)
{
  INDEX xbbl;
  INDEX iins;
  int i;
  BOOL change;
    
  xbbl = INVALID;
  /*
   * insert appropriate move instructions into xbbl
   */

  change = 1;

  while (change)
    {
      change = 0;
      for (i=0; i<64; i++)
	{
	  if (VALID(InSubst(i)) && !VALID(InSubst(InSubst(i))))
	    {
	      if (!VALID(xbbl))
		{
		  xbbl = InsertBblBeforeFancy(jbbl);
		}
	      
	      iins = AinsNew();
	      if (REG_IS_INT(i))
		AinsMakeLda(iins, InSubst(i), i, 0);
	      else 
		AinsMakeCpys(iins, InSubst(i),InSubst(i),i);
	      AINS_BBL(iins) = xbbl;
	      AinsPrepend(iins);
	      InSubst(i)=INVALID;
	      change = 1;
	    }
	}
    }


  if (VALID(xbbl)) ABBL_NINS(xbbl) = AbblNInss(xbbl);
  /*  if (VALID(xbbl) && jbbl<=1000)
    {
      printf("preadded\n");
      PrintBblSuper(stdout,xbbl);
      }*/

  /*
   * insert appropriate move instructions into xbbl
   */
  
  xbbl = INVALID;
  
  change = 1;
  
  while (change)
    {
      change = 0;
      for (i=0; i<64; i++)
	{
	  if (VALID(OutSubst(i)) && !VALID(OutSubst(OutSubst(i))))
	    {
	      if (!VALID(xbbl))
		{
		  xbbl = InsertBblAfterFancy(jbbl);
		}
	      
	      iins = AinsNew();
	      if (REG_IS_INT(i))
		AinsMakeLda(iins, i, OutSubst(i), 0);
	      else
		AinsMakeCpys(iins,i,i,OutSubst(i));
	      AINS_BBL(iins) = xbbl;
	      AinsAppend(iins);
	      OutSubst(i)=INVALID;
	      change = 1;
	    }
	}
    }
  
  if (VALID(xbbl)) ABBL_NINS(xbbl) = AbblNInss(xbbl);
  /*  if (VALID(xbbl) && jbbl<=1000)
    {
      printf("postadded\n");
      PrintBblSuper(stdout,xbbl);
      }*/

}
