#include "blocks.h"
#include "global.h" 
#include "functions.h"
#include "opcode.h"
#include "regs.h"
#include "flags.h"
#include "fltconst.h"

#define ABBL_IS_REACHABLE(i) ABBL_U3(i).doubleword.d1

extern void MarkLiveCode();
void  OptDataMarkLiveHellCode();
static void OptDataRelocFixpoint();
static void OptDataFinalize();

int inoptdata;
static int data_ins_change;

/********************************************************
 print the status of a data block
 ********************************************************/
void AblockPrintStatus(MAP_BLOCK * block)
{
  int tel;
  printf("block starting at address %lx with size %lx has status %x and the map looks like\n",block->start,block->size,block->status);
  for (tel=0;tel<block->size;tel+=8)
    printf("%x",block->map[tel/8]);
  printf("\n");
  for (tel=0;tel<block->size;tel+=8)
    printf("%x",block->otf_map[tel/8]);
  printf("\n");
}

/********************************************************
 return whether the data at an address is constant or not
 ********************************************************/

extern int AddressIsConstant(ADDRESS address)
{
  MAP_BLOCK * block = LocateBlock(address);

  if (!block) return 0;

  if (block->status & IS_DIRTY_STORE)
    return 0;

  if (block->map) 
    {
      if (!(block->map[(address-block->start)/8] & HAS_STORE))
	return 1;
    }
  else
    {
      if (!(block->status & HAS_STORE))
	return 1;
    }

  return 0;
}

/********************************************************
 do the same for the on-the-fly detection during CP
 ********************************************************/

extern int AddressIsConstantOTF(ADDRESS address)
{
  MAP_BLOCK * block = LocateBlock(address);

  if (!block) return 0;

  if (block->otf_status & IS_DIRTY_STORE)
    return 0;

  if (block->otf_map) 
    {
      if (!(block->otf_map[(address-block->start)/8] & HAS_STORE))
	{
	  //	  printf("considering %lx as loadable\n",address);
	  return 1;
	}
    }
  else
    {
      if (!(block->otf_status & HAS_STORE))
	{
	  //	  printf("considering %lx as loadable\n",address);
	  return 1;
	}
    }

  return 0;
}

/********************************************************
 adapt the block map when things happen at addresses
 ********************************************************/

extern void AdaptMap(int function,INT64 address)
{
  MAP_BLOCK * block=LocateBlock(address);
  int tel;
  if (!block) return;

  switch (function)
    {
    case CONST_LOAD: 
      if (block->map) 
	{
	  block->map[(address-block->start)/8] |=HAS_LOAD;
	  block->map[(address-block->start)/8] &= block->otf_map[(address-block->start)/8]|IS_JUMP_TARGET;
	}
      block->status|=HAS_LOAD; 
      break;
    case CONST_STORE: 
      if (block->map) 
	{
	  block->map[(address-block->start)/8] |=HAS_STORE;
	  block->map[(address-block->start)/8] &= block->otf_map[(address-block->start)/8]|IS_JUMP_TARGET;
	}
      block->status|=HAS_STORE; 
      break;
    case BLOCK_LOAD: 
      if (block->map && (!(block->status & IS_DIRTY_LOAD)) ) 
	for (tel=0;tel<block->size;tel+=8)
	  {
	    block->map[tel/8]|=HAS_LOAD|IS_DIRTY_LOAD;
	    block->map[tel/8] &= block->otf_map[tel/8]|IS_JUMP_TARGET;
	  }
      block->status|=HAS_LOAD|IS_DIRTY_LOAD; 
      break;
    case BLOCK_STORE: 
      if (block->map && (!(block->status & IS_DIRTY_STORE)) ) 
	for (tel=0;tel<block->size;tel+=8)
	  {
	    block->map[tel/8]|=HAS_STORE|IS_DIRTY_STORE;
	    block->map[tel/8] &= block->otf_map[tel/8]|IS_JUMP_TARGET;
	  }
      block->status|=HAS_STORE|IS_DIRTY_STORE; 
      break;
    }

  block->status &= block->otf_status|IS_DONE;

}

extern void AdaptMapOTF(int function,INT64 address,INDEX ibbl);

/********************************************************
 a block with a load instruction assumed that the load
 loaded a constant: add it to a linked list for the
 case it later turns out that it was not a constant
 ********************************************************/

extern void OTF_add(MAP_BLOCK * block,INT64 address,INDEX ibbl)
{
  int tel;
  OTF_INFO* tmp2;

  if (block->otf_map) 
    {
      OTF_INFO * tmp;
      if (!(block->otf))
	{
	  //	    if (((block->size/8)*8)!=block->size) printf("Oeps\n");
	  block->otf=(OTF_INFO**) malloc(sizeof(OTF_INFO *)*(block->size/8+1));
	  for (tel=0;tel<block->size;tel+=8)
	    block->otf[tel/8]=NULL;
	}
      tmp2 = block->otf[(address-block->start)/8];
      while (tmp2)
	{
	  if (tmp2->ibbl==ibbl) return;
	  tmp2=tmp2->next;
	}
      tmp=(OTF_INFO*) malloc(sizeof(OTF_INFO));
      tmp->next=block->otf[(address-block->start)/8];
      tmp->ibbl=ibbl;
      block->otf[(address-block->start)/8]=tmp;
    }
  else
    {
      printf("No map !!!!!\n");
    }
}

/********************************************************
 there is a store at an address detected during CP:
 all the bbls's that are in the linked list at that block
 need repropagation since they loaded what they assumed
 was a constant
 ********************************************************/

extern void OTF_mark(MAP_BLOCK * block,INT64 address)
{
  INT64 address2;
  if (block->otf_map) 
    {
      OTF_INFO * tmp,* tmp2;
      if (block->otf_map[(address-block->start)/8] & IS_REFQUAD)
	{
	  address2=*((INT64 *) (block->data+((address&(~(0x7L)))-block->start)));
	  if (!AddressIsCode(address2))
	    AdaptMapOTF(BLOCK_STORE, address2, 0); 
	}
      if (!(block->otf))
	return;
      for (tmp=block->otf[(address-block->start)/8]; tmp!=NULL; tmp=tmp2)
	{
	  tmp2=tmp->next;
	  AbblReprop(tmp->ibbl);
	  free(tmp);
	}
      block->otf[(address-block->start)/8]=NULL;
    }
  else
    {
      printf("No map !!!!!\n");
    }
}

/********************************************************
 adapt the block map when things happen at addresses during
 CP
 ********************************************************/
#define AEDG_LIVE(ibbl)      (AEDG_FLAGS(ibbl) &  FE_LIVE)
#define AEDG_UNLIVE(ibbl)     AEDG_FLAGS(ibbl) &= ~FE_LIVE
#define AEDG_SET_LIVE(ibbl)   AEDG_FLAGS(ibbl) |= FE_LIVE
extern int super_change;

extern void AdaptMapOTF(int function,INT64 address,INDEX ibbl)
{
  ADDRESS address2;
  MAP_BLOCK * block=LocateBlock(address);
  int tel = 0;

  if (inoptdata) return;

  if (!block) return;

  switch (function)
    {
    case CONST_LOAD: 
      if (block->otf_map)
	{
	  if (!(block->otf_map[(address-block->start)/8] & HAS_STORE))
	    OTF_add(block,address,ibbl);
	  block->otf_map[(address-block->start)/8] |=HAS_LOAD;
	  block->otf_map[(address-block->start)/8] &= block->map[(address-block->start)/8];
	  if (block->otf_map[(address-block->start)/8] & HAS_LOAD)
	    if (block->otf_map[(address-block->start)/8] & IS_REFQUAD)
	      {
		if (address2=ReadData64(address))
		  if (AddressIsCode(address2))
		    {
		      // set edge pointed to by address at location live
		      INDEX ifun=ABBL_FUN(AINS_BBL(Addr2Ins(address2)));
		      INDEX bbl=AFUN_BBL_FIRST(ifun);
		      INDEX iedg;
		      AFUN_FLAGS(ifun) |= F_MARKED;
		      FOREACH_BBL_PRED(bbl,iedg)
			if (ABBL_FUN(AEDG_HEAD(iedg))==AFUN_PSEUDO_HELL_LITE)
			  {
			    AEDG_MARK(iedg);
			    if (!AEDG_LIVE(iedg))
			      {
				super_change=1;
				//				printf("1 making edge to %d %s live\n",ifun,AFUN_NAME(ifun));
			      }
			    AEDG_SET_LIVE(iedg);
			  }
		    }
	      }
	}
      block->otf_status|=HAS_LOAD;
      break;
    case CONST_STORE: 
      if (block->otf_map) 
	{
	  block->otf_map[(address-block->start)/8] |=HAS_STORE;
	  block->otf_map[(address-block->start)/8] &= block->map[(address-block->start)/8];
	  if (block->otf_map[(address-block->start)/8] & (HAS_STORE|IS_DIRTY_STORE))
	    OTF_mark(block,address);
	}
      block->otf_status|=HAS_STORE;
      break;
    case BLOCK_STORE: 
      if (block->otf_map && (!(block->otf_status & IS_DONE)) ) 
	{
	  block->otf_status|=HAS_STORE|IS_DIRTY_STORE|HAS_LOAD|IS_DIRTY_LOAD|IS_DONE; 
	  for (tel=0;tel<block->size;tel+=8)
	    {
	      block->otf_map[tel/8]|=HAS_STORE|IS_DIRTY_STORE|HAS_LOAD|IS_DIRTY_LOAD;
	      block->otf_map[tel/8] &= block->map[tel/8];
	      if (block->otf_map[tel/8] & HAS_LOAD)
		if (block->otf_map[tel/8] & IS_REFQUAD)
		  {
		    if (address2=ReadData64(block->start+tel))
		      if (AddressIsCode(address2))
			{
			  // set edge pointed to by address at location live
			  
			  INDEX ifun=ABBL_FUN(AINS_BBL(Addr2Ins(address2)));
			  INDEX bbl=AFUN_BBL_FIRST(ifun);
			  INDEX iedg;
			  AFUN_FLAGS(ifun) |= F_MARKED;
			  FOREACH_BBL_PRED(bbl,iedg)
			    if (ABBL_FUN(AEDG_HEAD(iedg))==AFUN_PSEUDO_HELL_LITE)
			      {
				AEDG_MARK(iedg);
				if (!AEDG_LIVE(iedg))
				  {
				    super_change=1;
				    //				    printf("2 making edge to %d %s live\n",ifun,AFUN_NAME(ifun));
				  }
				AEDG_SET_LIVE(iedg);
			      }
			}
	      //	      if (block->otf_map[tel/8] & (HAS_STORE|IS_DIRTY_STORE))
		    OTF_mark(block,block->start+tel);
		  }
	    }
	}
      break;
    }
  
  block->otf_status &= block->status|IS_DONE;

}

/********************************************************
 empty the linked lists
 ********************************************************/

extern void OTF_cleanup()
{
  OTF_INFO * tmp,* tmp2;
  long tel,tel2;
  MAP_BLOCK * block;
  int teller = 0;

  for (tel=0; tel<numsections; tel++)
    {
      block = map_sections[tel].first;
      while (block)
	{
	  if (block->otf) 
	    {
	      for (tel2=0;tel2<block->size;tel2+=8)
		for (tmp=block->otf[tel2/8]; tmp!=NULL; tmp=tmp2)
		  {
		    tmp2=tmp->next;
		  
		    free(tmp);
		  }
	      free(block->otf);
	      block->otf=NULL;
	      
	      if (!(
		    R_SN_RCONST==map_sections[tel].type ||
		    R_SN_PDATA ==map_sections[tel].type ||
		    R_SN_RDATA ==map_sections[tel].type ||
		    R_SN_LITA  ==map_sections[tel].type ||
		    R_SN_LIT4  ==map_sections[tel].type ||
		    R_SN_LIT8  ==map_sections[tel].type
		    ))
		teller++;
	    }
	  block=block->next;
	}
    }
  //  printf("Found %d blocks that had data loaded that was extra read-only\n",teller);
}


#undef __PROC__
#define __PROC__ "OptDataInit"


/********************************************************
 initialize the data tags
 ********************************************************/

static void OptDataInit()
{

  long tel,tel2;
  MAP_BLOCK * block;

  for (tel=0; tel<numsections; tel++)
    {
      block = map_sections[tel].first;
      while (block)
	{
	  block->status=0;
	  if (block->map==NULL) 
	    {
	      //	      printf("No map in section %s\n",ASEC_NAME(BIN_SECTION_MAP[map_sections[tel].type]));
	      block=block->next;
	      continue;
	    }
	  for (tel2=0;tel2<block->size;tel2+=8)
	    block->map[tel2/8]&=IS_JUMP_TARGET;
	  block=block->next;
	}
    }

  SetEdgPropRegs();
}
/********************************************************
 walk over the instructions, stores and loads to constant
 addresses are detected and the maps are updated, 
 instructions that generate addresses have these addresses
 propagated in the program
 ********************************************************/

static void OptDataInsScan()
{
  INDEX iins;
  long base_value;
  ADDRESS address;
  MAP_BLOCK * block;
  INDEX ibbl;
  INDEX ifun;
  INDEX pred;
  BOOL flag;
  int skipped = 0;
  BOOL   verbose = HasOption(PRINT_OPT_VERBOSE,"Data");

  data_ins_change = 1;
  
  MarkLiveCode();
  
  while (data_ins_change)
    {
      skipped = 0;
      data_ins_change = 0;
      FOREACH_LIVE_INS(iins)
	{ 
	  if (!(ABBL_IS_REACHABLE(AINS_BBL(iins))))
	    {
	      skipped+=1;
	      continue;
	    }
	  if (AINS_IS_LOAD_STORE_TO_CONSTANT(iins))
	    {
	      if (AINS_REG_B(iins)==REG_GP)
		base_value=SysGetGlobalPointer();
	      else
		base_value=AINS_B_EVAL(iins);
	      address=base_value+AINS_DATA(iins);
	      
	      if (verbose)
		if (address == 0)
		  {
		    PrintIns(stdout,iins);
		    printf("Found zero load / store\n");
		    continue;
		  }
	      
	      AdaptMap(AINS_IS_LOAD(iins)?CONST_LOAD:CONST_STORE,address);
	}
	  
	  if (AINS_FLAGS(iins) & FI_EVAL_C)
	    {
	      address=AINS_C_EVAL(iins);
	      if (!AddressIsCode(address))
		{	  
		  block=LocateBlock(address);
		  if (block)
		    if (((block->status & 15)!= (IS_DIRTY_LOAD | IS_DIRTY_STORE | HAS_LOAD | HAS_STORE) ))
		      {
			AinsMarkUses(iins);
		      }
		}
	    }
	}

      OptDataMarkLiveHellCode();
      
    }
}

/********************************************************
 checking the tags in the blocks of the lita, a litamap 
 file is written and addresses are removed (well, made zero)
 ********************************************************/
static void OptDataLitaCleanup()
{
  long tel,tel2;
  ADDRESS address;
  MAP_BLOCK * block;
  char* litamapname =  (char *) malloc (256);
  FILE * litfile;
  MAP_BLOCK * block2;
  
  strcpy(litamapname,CONFIG_INFILE);
  strcat(litamapname,".litamap");


  if (!FACTOR_DONE)
    {
      litfile = fopen(litamapname,"w");
      
      for (tel=0;tel<numsections;tel++)
	{
	  if (map_sections[tel].type != R_SN_LITA) continue;
	  for (block=map_sections[tel].first; block!=NULL; block=block->next)
	    {
	      if ((block->status & ~IS_REFQUAD) == 0)
		{
		  block->attrib = -1;
		  fprintf(litfile,"0");
		  WriteData64(block->start,0L);
		}
	      else
		{
		  fprintf(litfile,"1");	  
		  address=ReadData64(block->start);
		  block2=LocateBlock(address);
		  if (block2)
		    {
		      if ((block2->status & IS_DONE2) == IS_DONE2) continue;
		      if (block2->map)
			{
			  for (tel2=0;tel2<block2->size;tel2+=8)
			    {
			      block2->map[tel2/8]|=HAS_LOAD|HAS_STORE|IS_DIRTY_STORE|IS_DIRTY_LOAD;
			      block2->map[tel2/8] &= block2->otf_map[tel2/8]|IS_JUMP_TARGET|IS_REFQUAD;
			    }
			}
		      else
			{
			  //			  printf("No map in Lita\n");
			}
		      block2->status|=HAS_LOAD|HAS_STORE|IS_DIRTY_STORE|IS_DIRTY_LOAD; // Veel te sterk, maar komt toch nooit voor ???
		      block2->status &= block2->otf_status|IS_DONE;
		    }
		}
	    }
	}
      
      fclose(litfile);
    }
  free(litamapname);
}

/********************************************************
 the refquad tags are set where necessary
 ********************************************************/
extern void Relocs2Map()
{
  INDEX isec;
  INDEX irel;
  ADDRESS address;
  MAP_BLOCK * block;
  FOREACH_SEC(isec)    
    {
      FOREACH_SEC_REL(isec,irel)
	{
	  if ( ASEC_AREL_TYPE(isec,irel) == R_REFQUAD)
	    {
	      if ( ASEC_AREL_EXTERN(isec,irel) )
		continue;
	      address=ASEC_AREL_ADDRESS(isec,irel);
	      block=LocateBlock(address);
	      block->map[(address-block->start)/8]|=IS_REFQUAD;
	      block->status|=IS_REFQUAD;
	      block->otf_map[(address-block->start)/8]|=IS_REFQUAD;
	      block->otf_status|=IS_REFQUAD;
	    }
	}
    }
}

/********************************************************
 a fixpoint is performed: if an address pointing to a block
 can be loaded, the data in that block can be loaded and
 written as well
 ********************************************************/


static void OptDataRelocFixpoint()
{
  register  BOOL change;
  long tel2,tel,tel3;
  ADDRESS address;
  MAP_BLOCK * block, *block2;
  do 
    {
      change=0;
      for (tel=0; tel<numsections; tel++)
	{
	  for (block=map_sections[tel].first; block!=NULL; block=block->next)
	    {
	      if (block->status&IS_DONE) continue;
	      if (!(block->status&HAS_LOAD)) continue;
	      if (!(block->status&IS_REFQUAD)) continue;
	      block->status|=IS_DONE;
	      if (block->map) 
		{
		  for (tel2=0;tel2<block->size;tel2+=8)
		    {
		      if (!(block->map[tel2/8]&HAS_LOAD)) continue;
		      if (!(block->map[tel2/8]&IS_REFQUAD)) continue;
		      address=*((INT64 *) (block->data+tel2));
		      block2=LocateBlock(address);
		      if (block2)
			{
			  if ((block2->status & IS_DONE2) == IS_DONE2) continue;
			  if ((block2->status & ~IS_REFQUAD)==0)
			    change=1;
			  block2->status|=HAS_LOAD|HAS_STORE|IS_DIRTY_LOAD|IS_DIRTY_STORE;
			  block2->status&=block2->otf_status|IS_DONE;
			  if (block2->map)
			    {
			      for (tel3=0;tel3<block2->size/8;tel3++)
				{
				  block2->map[tel3]|=HAS_LOAD|HAS_STORE|IS_DIRTY_STORE|IS_DIRTY_LOAD;
				  block2->map[tel3] &= block2->otf_map[tel3]|IS_JUMP_TARGET|IS_REFQUAD;
				}
			    }
			}
		    }

		}
	    }
	}
    } while (change);
}

/********************************************************
 for read-only sections, inaccuracies in the detection algorithms
 are masked
 ********************************************************/

extern long gap;

static void OptDataFinalize()
{

  int tel,tel2;
  MAP_BLOCK * block;

  for (tel=0; tel<numsections; tel++)
    {
      if (
	  R_SN_RCONST==map_sections[tel].type ||
	  R_SN_PDATA ==map_sections[tel].type ||
	  R_SN_RDATA ==map_sections[tel].type ||
	  R_SN_LITA  ==map_sections[tel].type ||
	  R_SN_LIT4  ==map_sections[tel].type ||
	  R_SN_LIT8  ==map_sections[tel].type
	  )
	{
	  for (block=map_sections[tel].first; block!=NULL; block=block->next)
	    { 
	      block->status &=~(IS_DIRTY_STORE|HAS_STORE);
	      for (tel2=0;tel2<block->size;tel2+=8)
		block->map[tel2/8] &=~(IS_DIRTY_STORE|HAS_STORE);
	    }
	}
      if ( R_SN_DATA==map_sections[tel].type)
	{
	  block=map_sections[tel].first;
	  while (block->start<0x140000000L+gap)
	    {
	      block->status &=~(IS_DIRTY_STORE|HAS_STORE);
	      for (tel2=0;tel2<block->size;tel2+=8)
		block->map[tel2/8] &=~(IS_DIRTY_STORE|HAS_STORE);
	      block=block->next;
	    }
	}
    }
}

/********************************************************
 a block that has no dirty loads: stores to addresses 
 from which no loads will be performed can be eliminated
 ********************************************************/

static void OptDataNoDirtyLoad2()
{
  long tel2;
  INDEX iins;
  long base_value;
  ADDRESS address;
  MAP_BLOCK* block;

  FOREACH_LIVE_INS(iins)
    {  
      if (AINS_IS_STORE_TO_CONSTANT(iins))
	{
	  if (AINS_REG_B(iins)==REG_GP)
	    base_value=SysGetGlobalPointer();
	  else
	    base_value=AINS_B_EVAL(iins);
	  address=base_value+AINS_DATA(iins);
	  block = LocateBlock(address);
	  if (block && !(block->status & IS_DIRTY_LOAD))
	    if (!(block->map[(address-block->start)/8]&HAS_LOAD))
	      {
		AinsKill(iins);
	      }
	  /*	    else
	      {
		if (AINS_FLAGS(iins) & FI_EVAL_A)
		  { 
		    printf("%lx is stored to %lx\n",AINS_A_EVAL(iins),address);
		    if (AddressIsCode(AINS_A_EVAL(iins))) printf("of fun %s\n",AFUN_NAME(ABBL_FUN(AINS_BBL(Addr2Ins(AINS_A_EVAL(iins))))));
		  }
	      }
	  */

	}
    }
}

static void OptDataNoDirtyLoad(MAP_BLOCK* block)
{
  long tel2;
  INDEX iins;
  long base_value;
  ADDRESS address;

  for (tel2=0;tel2<block->size;tel2+=8)
    if (!(block->map[tel2/8]&HAS_LOAD))
      {
	FOREACH_LIVE_INS(iins)
	  {  
	    if (AINS_IS_STORE_TO_CONSTANT(iins))
	      {
		if (AINS_REG_B(iins)==REG_GP)
		  base_value=SysGetGlobalPointer();
		else
		  base_value=AINS_B_EVAL(iins);
		address=base_value+AINS_DATA(iins);
		if (address==tel2+block->start)
		  {
		    AinsKill(iins);
		  }
	      }
	  }
      }
}
/********************************************************
 for the bss-section: if the only stores to an address
 store zero, they can be deleted
 ********************************************************/

static void OptDataNoDirtyLoadOrStore(long tel,MAP_BLOCK* block)
{
  INDEX iins;
  long base_value;
  ADDRESS address;

  if (ASEC_IS_BSS(BIN_SECTION_MAP[map_sections[tel].type])||ASEC_IS_SBSS(BIN_SECTION_MAP[map_sections[tel].type]))
    {
      int ok=1;
      FOREACH_LIVE_INS(iins)
	{
	  if (AINS_IS_STORE_TO_CONSTANT(iins))
	    {
	      if (AINS_REG_B(iins)==REG_GP)
		base_value=SysGetGlobalPointer();
	      else
		base_value=AINS_B_EVAL(iins);
	      address=base_value+AINS_DATA(iins);
	      if (address>=block->start && address<block->start+block->size)
		{
		  if (AINS_FLAGS(iins)&FI_EVAL_A)
		    {
		      if (AINS_A_EVAL(iins)!=0)
			ok=0;
		    }
		  else
		    ok=0;
		}
	    }
	  
	}
      if (ok==1)
	{
	  FOREACH_LIVE_INS(iins)
	    {
	      if (AINS_IS_STORE_TO_CONSTANT(iins))
		{
		  if (AINS_REG_B(iins)==REG_GP)
		    base_value=SysGetGlobalPointer();
		  else
		    base_value=AINS_B_EVAL(iins);
		  address=base_value+AINS_DATA(iins);
		  if (address>=block->start && address<block->start+block->size)
		    AinsKill(iins);
		}
	    }
	}
    }
}
/********************************************************
 blocks that are dead cannot have addresses in them loaded
 overwrite those with zeroes, for hell-edge removals
 ********************************************************/
  
static void OptDataNoLoad(long tel,MAP_BLOCK* block)
{

  INDEX iins;
  INDEX isec;
  ADDRESS address;
  long base_value;
  BOOL   verbose = HasOption(PRINT_OPT_VERBOSE,"Data");

  isec=BIN_SECTION_MAP[map_sections[tel].type];
  if (
      (VALID(isec)) && 
      (
       R_SN_DATA==map_sections[tel].type ||
       R_SN_RDATA==map_sections[tel].type ||
       R_SN_XDATA==map_sections[tel].type ||
       R_SN_PDATA==map_sections[tel].type ||
       R_SN_SDATA==map_sections[tel].type || 
       R_SN_RCONST==map_sections[tel].type
       )
      )
    {
      int new=0;
      long tel2;
      if (!(block->status&IS_DIRTY_STORE)) // GROTE VERANDERING: soms een dirty store die niet weg te krijgen is, maar geen load
	block->attrib = -1; 
      for (tel2=block->start ; tel2<block->start+block->size; tel2+=8) // OPPASSEN MET BLOKKEN DIE GEEN 8-voud zijn !!!!
	{
	  if ( ReadData64(tel2)!=0) new=1;
	  WriteData64(tel2,0);
	}
      if (verbose)
	if (new)
	  printf("found dead data           at %lx -> %lx %s!\n",block->start,block->start+block->size,ASEC_NAME(isec));

      if (block->status & HAS_STORE)
	{
	  FOREACH_LIVE_INS(iins)
	    {
	      if (AINS_IS_STORE_TO_CONSTANT(iins))
		{
		  if (AINS_REG_B(iins)==REG_GP)
		    base_value=SysGetGlobalPointer();
		  else
		    base_value=AINS_B_EVAL(iins);
		  address=base_value+AINS_DATA(iins);
		  if (address>=block->start && address<block->start+block->size)
		    AinsKill(iins);
		}
	    }
	}
    }
  else
    {
      if (
	  (VALID(isec)) && 
	  (
	   R_SN_BSS==map_sections[tel].type ||
	   R_SN_SBSS==map_sections[tel].type
	   )
	  )
	{
	  if (block->status & HAS_STORE)
	    {
	      FOREACH_LIVE_INS(iins)
		{
		  if (AINS_IS_STORE_TO_CONSTANT(iins))
		    {
		      if (AINS_REG_B(iins)==REG_GP)
			base_value=SysGetGlobalPointer();
		      else
			base_value=AINS_B_EVAL(iins);
		      address=base_value+AINS_DATA(iins);
		      if (address>=block->start && address<block->start+block->size)
			AinsKill(iins);
		    }
		}
	    }
	}
    }
}

/********************************************************
 set the necessary tags in the maps and that optimize for
 those tags 
 ********************************************************/

extern void OptData()
{
  long tel;
  MAP_BLOCK * block;
  BOOL   verbose = HasOption(PRINT_OPT_VERBOSE,"Data");
  long dirtytel;

  inoptdata=1;

  ResetAddressCalc();  

  InitMarking();

  /* Init */

  OptDataInit();

  /* Instruction scan */

  OptDataInsScan();

  if (verbose)
    {
      dirtytel=0L;
      for (tel=0; tel<numsections; tel++)
	{
	  block = map_sections[tel].first;
	  while (block)
	    {
	      if ((block->status&IS_DIRTY_LOAD)||(block->status&IS_DIRTY_STORE)) dirtytel++;
	      block=block->next;
	    }
	}
      printf("OptData Phase 1 %ld dirty\n",dirtytel);
    }
  
  /* Lita cleanup */

  OptDataLitaCleanup();

  if (verbose)
    {
      dirtytel=0L;
      for (tel=0; tel<numsections; tel++)
	{
	  block = map_sections[tel].first;
	  while (block)
	    {
	      if ((block->status&IS_DIRTY_LOAD)||(block->status&IS_DIRTY_STORE)) dirtytel++;
	      block=block->next;
	    }
	}
      printf("OptData Phase 2 %ld dirty\n",dirtytel);
    }
  
  /* Relocation fixpoint */

  Relocs2Map();
  OptDataRelocFixpoint();
  OptDataFinalize();

    
  if (verbose)
    {
      dirtytel=0L;
      for (tel=0; tel<numsections; tel++)
	{
	  block = map_sections[tel].first;
	  while (block)
	    {
	      if ((block->status&IS_DIRTY_LOAD)||(block->status&IS_DIRTY_STORE)) dirtytel++;
	      block=block->next;
	    }
	}
      printf("Phase 3 %ld dirty\n",dirtytel);
    }

  OptDataNoDirtyLoad2();

  for (tel=0; tel<numsections; tel++)
    {
      block = map_sections[tel].first;
      for (block=map_sections[tel].first; block!=NULL; block=block->next)
	{ 
	  if (!(block->status&HAS_LOAD))
	    {
	      OptDataNoLoad(tel,block);
	    }

	  if (!(block->status&HAS_STORE))
	    {
	      if (verbose)
		printf("found constant data       at %lx -> %lx!\n",block->start,block->start+block->size);

	      continue;
	    }

	  if (!(block->status & IS_DIRTY_LOAD))
	    {
	      //      OptDataNoDirtyLoad(block);
	    }
	  
	  if (!(block->status & IS_DIRTY_LOAD))
	    if (!(block->status & IS_DIRTY_STORE))
	      {
		if (verbose)
		  printf("found all load and stores at %lx -> %lx with size %10lx!\n",block->start,block->start+block->size,block->size);
		
		OptDataNoDirtyLoadOrStore(tel,block);

	      }
	}
    }

  ResetAddressCalc();
  GenerateMapFile();
  inoptdata=0;

}

#define ABBL_MAGIC(i) ABBL_U1(i).doubleword.d1

#undef __PROC__
#define __PROC__ "MakeReachableAddress"

extern void MakeReachableAddress(ADDRESS address)
{
  INDEX irel;
  MAP_BLOCK * block = LocateBlock(address);
  INDEX isec;
  if (!block) return;
  isec = LocateSection(address);
  if (!VALID(isec)) return;
  
  if( ASEC_IS_LITA(isec) ) return;
  
  FOREACH_SEC_REL(isec,irel)
    {
      if( ASEC_AREL_EXTERN(isec,irel) )
	continue;
      
      if(  ASEC_AREL_TYPE(isec,irel) == R_LITERAL )
	continue;
      
      if( ASEC_AREL_SECTION(isec,irel) != R_SN_TEXT &&
	  ASEC_AREL_SECTION(isec,irel) != R_SN_INIT &&
	  ASEC_AREL_SECTION(isec,irel) != R_SN_NULL &&
	  ASEC_AREL_SECTION(isec,irel) != R_SN_FINI )
	continue;

      if (ASEC_AREL_ADDRESS(isec,irel) == address)
	if (ASEC_AREL_TYPE(isec,irel)==R_REFQUAD)
	  if (ReadData64(ASEC_AREL_ADDRESS(isec,irel)))
	    {
	      if (ABBL_MAGIC(AINS_BBL(Addr2Ins(ReadData64(ASEC_AREL_ADDRESS(isec,irel)))))) continue;
	      ABBL_MAGIC(AINS_BBL(Addr2Ins(ReadData64(ASEC_AREL_ADDRESS(isec,irel))))) = 1;
	    }
    }
  return;
}

/*****************************************
 the rest looks like unused code
*****************************************/

#undef __PROC__
#define __PROC__ "MakeReachableBlock"

extern void MakeReachableBlock(ADDRESS address)
{
  INDEX irel;
  MAP_BLOCK * block = LocateBlock(address);
  INDEX isec;
  if (!block) return;
  isec = LocateSection(address);
  
  if (!VALID(isec)) return;

  if( ASEC_IS_LITA(isec) ) return;

  FOREACH_SEC_REL(isec,irel)
    {
      if( ASEC_AREL_EXTERN(isec,irel) )
	continue;
      
      if(  ASEC_AREL_TYPE(isec,irel) == R_LITERAL )
	continue;
      
      if( ASEC_AREL_SECTION(isec,irel) != R_SN_TEXT &&
	  ASEC_AREL_SECTION(isec,irel) != R_SN_INIT &&
	  ASEC_AREL_SECTION(isec,irel) != R_SN_NULL &&
	  ASEC_AREL_SECTION(isec,irel) != R_SN_FINI )
	{
	  if (LocateBlock(ASEC_AREL_ADDRESS(isec,irel)) == block)
	    if (ASEC_AREL_TYPE(isec,irel)==R_REFQUAD)
	      if (ReadData64(ASEC_AREL_ADDRESS(isec,irel)))
		MakeReachableAddress(ReadData64(ASEC_AREL_ADDRESS(isec,irel)));
	  continue;
	}
      
      if (LocateBlock(ASEC_AREL_ADDRESS(isec,irel)) == block)
	if (ASEC_AREL_TYPE(isec,irel)==R_REFQUAD)
	  if (ReadData64(ASEC_AREL_ADDRESS(isec,irel)))
	    {
	      if (ABBL_MAGIC(AINS_BBL(Addr2Ins(ReadData64(ASEC_AREL_ADDRESS(isec,irel)))))) continue;
	      ABBL_MAGIC(AINS_BBL(Addr2Ins(ReadData64(ASEC_AREL_ADDRESS(isec,irel))))) = 1;
	    }
    }
  return;
}

/********************************************************
 walk over the instructions, stores and loads to constant
 addresses are detected and the maps are updated, 
 instructions that generate addresses have these addresses
 propagated in the program
 ********************************************************/

extern void OptDataInsScanOTF()
{
  INDEX iins;
  long base_value;
  ADDRESS address;
  MAP_BLOCK * block;

  FOREACH_LIVE_INS(iins)
    {
      if (AINS_IS_LOAD_STORE_TO_CONSTANT(iins))
	{
	  if (AINS_REG_B(iins)==REG_GP)
	    base_value=SysGetGlobalPointer();
	  else
	    base_value=AINS_B_EVAL(iins);
	  address=base_value+AINS_DATA(iins);

	  block=LocateBlock(address);
	  if (block)
	    {
	      if (AINS_IS_LOAD(iins))
		{
		  if (block->otf_map) 
		    block->otf_map[(address-block->start)/8]|=HAS_LOAD;
		  block->otf_status|=HAS_LOAD;
		}
	      else
		{
		  if (block->otf_map) 
		    block->otf_map[(address-block->start)/8]|=HAS_STORE;
		  block->otf_status|=HAS_STORE;
		}
	    }
	}
    }
}


extern void MarkLiveCode()
{
  BOOL   change = 0;
  INDEX iedg;
  INDEX ibbl;
  INDEX ifun;
  INDEX iins;
  BOOL  flag;
  BOOL  flag2;
  NUMBER rcount;
  NUMBER icount;
  NUMBER bcount;
  NUMBER fcount;
  clock_t time = clock();
    
  FOREACH_LIVE_BBL(ibbl)
    {
      if( AFUN_IS_PSEUDO(ABBL_FUN(ibbl)) && (ABBL_FUN(ibbl) != AFUN_PSEUDO_HELL_LITE))
	{
	  ABBL_MARK(ibbl);
	  ABBL_IS_REACHABLE(ibbl) = 1;
	}
      else
	{
	  ABBL_UNMARK(ibbl);
	  ABBL_IS_REACHABLE(ibbl) = 0;
	}
      
      if (ABBL_FLAGS(ibbl)&FB_PROG_ENTRY)
	{
	  ABBL_MARK(ibbl);
	  ABBL_IS_REACHABLE(ibbl) = 1;
	}
    }

  FOREACH_LIVE_FUN(ifun)
    {
      if( AFUN_IS_PSEUDO(ifun) && ifun != AFUN_PSEUDO_HELL_LITE)
	AFUN_FLAGS(ifun)|=F_MARKED;
      else
	AFUN_FLAGS(ifun)&=~F_MARKED;
    }
  
  do
    {
      flag = 0;
      FOREACH_LIVE_FUN(ifun)
	{
	  if (ifun==AFUN_PSEUDO_HELL_LITE) continue;
	  if( !(AFUN_FLAGS(ifun)&F_MARKED) )
	    continue;
	  
	  AFUN_FLAGS(ifun)&=~F_MARKED;
	  
	  do
	    {
	      flag2 = 0;
	      
	      FOREACH_FUN_BBL(ifun,ibbl)
		{
		  if (!ABBL_IS_MARKED(ibbl)) continue;
		  
		  ABBL_UNMARK(ibbl);
		  
		  FOREACH_BBL_SUCC(ibbl,iedg)
		    {
		      INDEX jbbl = AEDG_TAIL(iedg);
		      
		      /* do not follow return edges,
			 the link edge will guarantee the survival */
		      
		      if( AEDG_TYPE(iedg) == ET_FUN_RETURN )
			continue;
		      
		      if( ABBL_IS_REACHABLE(jbbl) == 0 )
			{
			  ABBL_IS_REACHABLE(jbbl) = 1;
			  ABBL_MARK(jbbl);
			  if (!AEDG_IS_INTER(iedg))
			    {
			      flag2 = 1;
			    }
			  else
			    {
			      AFUN_FLAGS(ABBL_FUN(jbbl))|=F_MARKED;
			      flag = 1;
			    }
			}
		    }
		}
	    }
	  while (flag2);
	}
    }
  while (flag);

}

void OptDataMarkLiveHellCode()
{
  BOOL change = 0;
  INDEX irel;
  INDEX iins;
  INDEX iedg,ibbl;
  NUMBER count = 0;
  INDEX isec;
  INDEX ifun;
  MAP_BLOCK * block;
  BOOL flag, flag2;
  int teller1,teller2;

  Relocs2Map();
  OptDataRelocFixpoint();

  FOREACH_LIVE_FUN(ifun)
    AFUN_FLAGS(ifun)&=~F_MARKED;

  FOREACH_LIVE_BBL(ibbl)
    ABBL_UNMARK(ibbl);

  /* Mark functions who's address is loaded from the lita */

  teller1=0;

  FOREACH_LIVE_INS(iins)
    {
      if (!ABBL_IS_REACHABLE(AINS_BBL(iins))) continue;
      if (AINS_FLAGS(iins) & FI_EVAL_C)
	{
	  if (AINS_REG_C(iins)==REG_ZERO ) continue;
	  if (AddressIsCode(AINS_C_EVAL(iins)))
	    {
	      if (AINS_TYPE(iins) == IT_JMP || AINS_TYPE(iins) == IT_BRA) continue;
	      if (!(AFUN_FLAGS(ABBL_FUN(AINS_BBL(Addr2Ins(AINS_C_EVAL(iins))))) & F_KILLED))
		{
		  AFUN_FLAGS(ABBL_FUN(AINS_BBL(Addr2Ins(AINS_C_EVAL(iins))))) |= F_MARKED;
		  if (!ABBL_IS_REACHABLE(AINS_BBL(Addr2Ins(AINS_C_EVAL(iins)))))
		    {
		      ABBL_IS_REACHABLE(AINS_BBL(Addr2Ins(AINS_C_EVAL(iins)))) = 1;
		      data_ins_change = 1;
		      teller1++;
		    }
		}
	    }
	}
    }

  teller2=0;

  FOREACH_SEC(isec)    /* Mark functions who's address is found in other data sections */
    {
      if( ASEC_IS_LITA(isec) ) continue;

      FOREACH_SEC_REL(isec,irel)
	{
	  if( ASEC_AREL_EXTERN(isec,irel) )
	    continue;
	  
	  if(  ASEC_AREL_TYPE(isec,irel) == R_LITERAL )
	    continue;
	  
	  if( ASEC_AREL_SECTION(isec,irel) != R_SN_TEXT &&
	      ASEC_AREL_SECTION(isec,irel) != R_SN_INIT &&
	      ASEC_AREL_SECTION(isec,irel) != R_SN_NULL &&
	      ASEC_AREL_SECTION(isec,irel) != R_SN_FINI )
	    continue;
	  
	  if (ASEC_AREL_TYPE(isec,irel)==R_REFQUAD)
	    {
	      block = LocateBlock(ASEC_AREL_ADDRESS(isec,irel));
	      if (!(block->map[(ASEC_AREL_ADDRESS(isec,irel)-block->start)/8] & HAS_LOAD) && block->status==0) continue;
	      if (ReadData64(ASEC_AREL_ADDRESS(isec,irel)))
		{
		  AFUN_FLAGS(ABBL_FUN(AINS_BBL(Addr2Ins(ReadData64(ASEC_AREL_ADDRESS(isec,irel)))))) |= F_MARKED;
		  ABBL_MARK(AINS_BBL(Addr2Ins(ReadData64(ASEC_AREL_ADDRESS(isec,irel)))));
		  if (!ABBL_IS_REACHABLE(AINS_BBL(Addr2Ins(ReadData64(ASEC_AREL_ADDRESS(isec,irel))))))
		    {	
		      ABBL_IS_REACHABLE(AINS_BBL(Addr2Ins(ReadData64(ASEC_AREL_ADDRESS(isec,irel)))))=1;
		      data_ins_change = 1;
		      teller2++;
		    }
		}
	    }
	}
    }

  /* mark the following the basic blocks as alive */

  FOREACH_LIVE_BBL(ibbl)
    if (ABBL_IS_REACHABLE(ibbl))
      {
	ABBL_MARK(ibbl);
	AFUN_FLAGS(ABBL_FUN(ibbl))|=F_MARKED;
      }
  do
    {
      flag = 0;
      FOREACH_LIVE_FUN(ifun)
	{
	  if (ifun==AFUN_PSEUDO_HELL_LITE) continue;
	  if( !(AFUN_FLAGS(ifun)&F_MARKED) )
	    continue;
	  
	  AFUN_FLAGS(ifun)&=~F_MARKED;
	
	  //	  printf("fun %s %d is added\n",AFUN_NAME(ifun),ifun);
  
	  do
	    {
	      flag2 = 0;
	      
	      FOREACH_FUN_BBL(ifun,ibbl)
		{
		  if (!ABBL_IS_MARKED(ibbl)) continue;
		  if (!ABBL_IS_REACHABLE(ibbl)) continue;
		  
		  ABBL_UNMARK(ibbl);
		  
		  FOREACH_BBL_SUCC(ibbl,iedg)
		    {
		      INDEX jbbl = AEDG_TAIL(iedg);
		      
		      /* do not follow return edges,
			 the link edge will guarantee the survival */
		      
		      if( AEDG_TYPE(iedg) == ET_FUN_RETURN )
			continue;
		      
		      if( ABBL_IS_REACHABLE(jbbl) == 0 )
			{
			  ABBL_IS_REACHABLE(jbbl) = 1;
			  ABBL_MARK(jbbl);
			  if (!AEDG_IS_INTER(iedg))
			    {
			      flag2 = 1;
			    }
			  else
			    {
			      AFUN_FLAGS(ABBL_FUN(jbbl))|=F_MARKED;;
			      flag = 1;
			    }
			}
		    }
		}
	    }
	  while (flag2);
	}
    }
  while (flag);

  if (!data_ins_change)
    FOREACH_LIVE_FUN(ifun)
      if (!ABBL_IS_REACHABLE(AFUN_BBL_FIRST(ifun)))
	FOREACH_BBL_PRED(AFUN_BBL_FIRST(ifun),iedg)
	{
          if( AEDG_TYPE(iedg)==ET_HELL)
	    {
	      //		printf("remove hell edge to procedure %d %s\n",ifun,AFUN_NAME(ifun));
		AedgKillFancy(iedg);
              count++;
	    }
	}
}
